/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"
#include <memory>
#include "common/common.h"
#include "defs.h"
#include "errors.h"
#include "parser/ast.h"
#include "spdlog/spdlog.h"

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query 
 */
std::shared_ptr<Query> Analyze::do_analyze(
    std::shared_ptr<ast::TreeNode> parse) {
    std::shared_ptr<Query> query = std::make_shared<Query>();
    if (auto select_stmt = std::dynamic_pointer_cast<ast::SelectStmt>(parse)) {
        // 处理表名
        query->tables = std::move(select_stmt->tabs);
        // 检查表是否存在
        for (auto& tb : query->tables) {
            if (!sm_manager_->db_.is_table(tb)) {
                spdlog::error("[do_analyze] table {} not found", tb);
                throw TableNotFoundError(tb);
            }
        }

        // 处理target list，再target list中添加上表名，例如 a.id
        // select id from a; , select_stmt->cols = {a.id}
        // select count(a.id) as cnt from a; , select_stmt->cols[0] = {table=a, col=id, alias=cnt, aggr_type=AGG_COUNT}
        for (auto& sv_sel_col : select_stmt->cols) {
            TabCol sel_col = {.tab_name = sv_sel_col->tab_name,
                              .col_name = sv_sel_col->col_name,
                              .alias = sv_sel_col->alias,
                              .aggr_type = sv_sel_col->aggr_type};
            query->cols.push_back(sel_col);
        }

        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);
        if (query->cols.empty()) {
            // select all columns
            for (auto& col : all_cols) {
                TabCol sel_col = {.tab_name = col.tab_name,
                                  .col_name = col.name,
                                  .alias = "",
                                  .aggr_type = col.aggr_type};
                query->cols.push_back(sel_col);
            }
        } else {
            // infer table name from column name
            for (auto& sel_col : query->cols) {
                sel_col = check_column(all_cols, sel_col);  // 列元数据校验
            }
        }
        //处理where条件
        get_clause(select_stmt->conds, query->conds);
        check_clause(query->tables, query->conds, false);

        // 处理group by
        if (select_stmt->groupby != nullptr) {
            for (auto& sv_group_col : select_stmt->groupby->cols) {
                TabCol group_col = {.tab_name = sv_group_col->tab_name,
                                    .col_name = sv_group_col->col_name,
                                    .alias = sv_group_col->alias,
                                    .aggr_type = sv_group_col->aggr_type};
                group_col = check_column(all_cols, group_col);  // 补全表名
                query->groupby_cols.push_back(group_col);
            }
            get_clause(select_stmt->groupby->conds, query->having_conds);
            check_clause(query->tables, query->having_conds, true);

            // 校验：select中是否出现过不在group by中的普通col
            for (auto& sel_col : query->cols) {
                if (sel_col.aggr_type != AGG_NULL) {
                    continue;
                }
                bool in_groupby = false;
                for (auto& group_col : query->groupby_cols) {
                    if (sel_col.tab_name == group_col.tab_name &&
                        sel_col.col_name == group_col.col_name) {
                        in_groupby = true;
                        break;
                    }
                }
                if (!in_groupby) {
                    throw AggregationError(
                        sel_col.tab_name + '.' + sel_col.col_name +
                        " is not in group by and is not aggregated");
                }
            }
        }

        // 校验：select中是否有普通col 和 聚集col 同时存在的情况，且没有 group by 部分
        if (query->groupby_cols.empty()) {
            if (select_stmt->has_aggr) {
                for (auto& sel_col : query->cols) {
                    if (sel_col.aggr_type == AGG_NULL) {
                        throw AggregationError(
                            "Not having a group_by clause, but both aggr and "
                            "non-aggr col exist");
                    }
                }
            }
        }

    } else if (auto update_stmt =
                   std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
        /** 
         * 1. 检验 set_clauses
         *   1.1 检验字段名是否合法，塞入 Query
         *   1.2 转换 Value
         * 2. 检验where条件 
         */
        std::vector<ColMeta> all_cols;
        get_all_cols({update_stmt->tab_name}, all_cols);
        for (auto& set_clause : update_stmt->set_clauses) {
            TabCol set_col;
            set_col.col_name = set_clause->col_name;
            set_col.tab_name = update_stmt->tab_name;
            set_col = check_column(all_cols, set_col);
            query->set_clauses.push_back(
                {set_col, convert_sv_value(set_clause->val)});
        }
        get_clause(update_stmt->conds, query->conds);
        check_clause({update_stmt->tab_name}, query->conds, false);

    } else if (auto delete_stmt =
                   std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
        //处理where条件
        get_clause(delete_stmt->conds, query->conds);
        check_clause({delete_stmt->tab_name}, query->conds, false);
    } else if (auto insert_stmt =
                   std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
        // 处理insert 的values值
        for (auto& sv_val : insert_stmt->vals) {
            query->values.push_back(convert_sv_value(sv_val));
        }
    } else {
        // do nothing
    }
    query->parse = std::move(parse);
    return query;
}

TabCol Analyze::check_column(const std::vector<ColMeta>& all_cols,
                             TabCol target) {
    if (target.col_name == "*" && target.aggr_type == AGG_COUNT) {
        return target;
    }
    if (target.tab_name.empty()) {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto& col : all_cols) {
            if (col.name == target.col_name) {
                if (!tab_name.empty()) {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty()) {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    } else {
        /** 保证目标列存在 */
        // a.id
        // a 这个表中是否存在 id 字段
        if (!sm_manager_->db_.get_table(target.tab_name)
                 .is_col(target.col_name)) {
            throw ColumnNotFoundError(target.tab_name + '.' + target.col_name);
        }
    }
    return target;
}

void Analyze::get_all_cols(const std::vector<std::string>& tab_names,
                           std::vector<ColMeta>& all_cols) {
    for (auto& sel_tab_name : tab_names) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto& sel_tab_cols =
            sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(),
                        sel_tab_cols.end());
    }
}

void Analyze::get_clause(
    const std::vector<std::shared_ptr<ast::BinaryExpr>>& sv_conds,
    std::vector<Condition>& conds) {
    // 处理where条件：从ast中获取条件表达式，将表达式中的lhs、rhs、op转换成Condition结构
    conds.clear();  // 保证conds为空，随后往里添加
    for (auto& expr : sv_conds) {
        Condition cond;
        cond.lhs_col = {.tab_name = expr->lhs->tab_name,
                        .col_name = expr->lhs->col_name,
                        .alias = expr->lhs->alias,
                        .aggr_type = expr->lhs->aggr_type};
        cond.op = convert_sv_comp_op(expr->op);
        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col =
                       std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name,
                            .col_name = rhs_col->col_name,
                            .alias = rhs_col->alias,
                            .aggr_type = rhs_col->aggr_type};
        }
        conds.push_back(cond);
    }
}

// 1. 获取表达式中的可能得隐式表名，同时检查列的合法性
// 2. 转换查询条件，对每个字段进行类型检查
// 3. 类型比较不一致抛异常
void Analyze::check_clause(const std::vector<std::string>& tab_names,
                           std::vector<Condition>& conds, bool is_having) {
    // auto all_cols = get_all_cols(tab_names);
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto& cond : conds) {
        // Infer table name from column name
        if (!is_having && cond.lhs_col.aggr_type != AGG_NULL) {
            throw AggregationError(
                "Aggregation cannot be used in WHERE clause");
        }
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if (!cond.is_rhs_val) {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }

        if (cond.lhs_col.aggr_type == AGG_COUNT &&
            cond.lhs_col.col_name == "*") {
            ColType lhs_type = TYPE_INT, rhs_type;
            if (cond.is_rhs_val) {
                cond.rhs_val.init_raw(sizeof(int));
                rhs_type = cond.rhs_val.type;
            } else {
                TabMeta& rhs_tab =
                    sm_manager_->db_.get_table(cond.rhs_col.tab_name);
                auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
                rhs_type = rhs_col->type;
            }
            if (lhs_type != rhs_type &&
                !(lhs_type == TYPE_INT && rhs_type == TYPE_FLOAT)) {
                throw IncompatibleTypeError(coltype2str(lhs_type),
                                            coltype2str(rhs_type));
            }
            continue;
        }

        TabMeta& lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            // cond.rhs_val.init_raw(lhs_col->len);
            // 进行类型转换并初始化
            cond.rhs_val.type_convert_init(*lhs_col);
            rhs_type = cond.rhs_val.type;
        } else {
            TabMeta& rhs_tab =
                sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        }
        if (lhs_type != rhs_type) {
            throw IncompatibleTypeError(coltype2str(lhs_type),
                                        coltype2str(rhs_type));
        }
    }
}

Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value>& sv_val) {
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
        val.set_int(int_lit->val);
    } else if (auto float_lit =
                   std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
        val.set_float(float_lit->val);
    } else if (auto str_lit =
                   std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
        val.set_str(str_lit->val);
    } else {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
    std::map<ast::SvCompOp, CompOp> m = {
        {ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
        {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
    };
    return m.at(op);
}
