/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cfloat>
#include <climits>
#include <cstddef>
#include <cstring>
#include <memory>
#include <string>
#include <vector>
#include "defs.h"
#include "parser/ast.h"
#include "record/rm_defs.h"
#include "system/sm_meta.h"

struct TabCol {
    std::string tab_name;
    std::string col_name;
    std::string alias;
    AggrType aggr_type = AggrType::AGG_NULL;

    friend bool operator<(const TabCol& x, const TabCol& y) {
        return std::make_pair(x.tab_name, x.col_name) <
               std::make_pair(y.tab_name, y.col_name);
    }
};

struct Value {
    ColType type;  // type of value
    union {
        int int_val;      // int value
        float float_val;  // float value
    };
    std::string str_val;  // string value

    std::shared_ptr<RmRecord> raw;  // raw record buffer

    void set_int(int int_val_) {
        type = TYPE_INT;
        int_val = int_val_;
    }

    void set_float(float float_val_) {
        type = TYPE_FLOAT;
        float_val = float_val_;
    }

    void set_str(std::string str_val_) {
        type = TYPE_STRING;
        str_val = std::move(str_val_);
    }

    void init_raw(int len) {
        assert(raw == nullptr);
        raw = std::make_shared<RmRecord>(len);
        if (type == TYPE_INT) {
            assert(len == sizeof(int));
            *(int*)(raw->data) = int_val;
        } else if (type == TYPE_FLOAT) {
            assert(len == sizeof(float));
            *(float*)(raw->data) = float_val;
        } else if (type == TYPE_STRING) {
            if (len < (int)str_val.size()) {
                throw StringOverflowError();
            }
            memset(raw->data, 0, len);
            memcpy(raw->data, str_val.c_str(), str_val.size());
            if (static_cast<int>(str_val.size()) < len) {
                raw->data[str_val.size()] = '\0';
            }
            // printf("init_raw: %s\n", str_val.c_str());
        }
    }

    void type_convert_init(const ColMeta& another_col) {
        // ! check: maybe add more type conversion here
        // should combine init_raw() and this function
        int len = 0;
        if (type == TYPE_INT && another_col.type == TYPE_FLOAT) {
            set_float((float)int_val);
        } else if (type == TYPE_FLOAT && another_col.type == TYPE_INT) {
            set_int((int)float_val);
        }

        if (type == TYPE_INT) {
            len = sizeof(int);
        } else if (type == TYPE_FLOAT) {
            len = sizeof(float);
        } else if (type == TYPE_STRING) {
            len = another_col.len;
            if (len < (int)str_val.size()) {
                throw StringOverflowError();
            }
        }
        init_raw(len);
    }

    void init_max(int len) {
        if (type == TYPE_INT) {
            set_int(INT_MAX);
            init_raw(sizeof(int));
        } else if (type == TYPE_FLOAT) {
            set_float(FLT_MAX);
            init_raw(sizeof(float));
        } else if (type == TYPE_STRING) {
            std::string str(len, '\xff');
            set_str(str);
            init_raw(len);
        }
    }
    void init_min(int len) {
        if (type == TYPE_INT) {
            set_int(INT_MIN);
            init_raw(sizeof(int));
        } else if (type == TYPE_FLOAT) {
            set_float(FLT_MIN);
            init_raw(sizeof(float));
        } else if (type == TYPE_STRING) {
            std::string str(len, '\0');
            set_str(str);
            init_raw(len);
        }
    }
    void init_zero() {
        if (type == TYPE_INT) {
            set_int(0);
            init_raw(sizeof(int));
        } else if (type == TYPE_FLOAT) {
            set_float(0.0);
            init_raw(sizeof(float));
        } else if (type == TYPE_STRING) {
            set_str("");
            init_raw(0);
        }
    }
};

enum CompOp { OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_GE };

struct Condition {
    TabCol lhs_col;   // left-hand side column
    CompOp op;        // comparison operator
    bool is_rhs_val;  // true if right-hand side is a value (not a column)
    TabCol rhs_col;   // right-hand side column
    Value rhs_val;    // right-hand side value
};

struct SetClause {
    TabCol lhs;
    Value rhs;
};
