/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstring>
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_seq_scan.h"
#include "index/ix.h"
#include "index/ix_index_handle.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class SortExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;
    std::vector<ColMeta> cols_;
    ColMeta order_col;
    size_t tuple_num;
    size_t tuple_idx;
    bool is_desc_;
    std::vector<size_t> used_tuple;
    std::unique_ptr<RmRecord> current_tuple;
    std::vector<std::unique_ptr<RmRecord>> records;

   public:
    SortExecutor(std::unique_ptr<AbstractExecutor> prev, TabCol sel_cols,
                 bool is_desc) {
        prev_ = std::move(prev);
        cols_ = prev_->cols();
        order_col = get_col_offset(sel_cols);
        is_desc_ = is_desc;
        tuple_num = 0;
        used_tuple.clear();
        records.clear();
    }

    void beginTuple() override {
        tuple_idx = 0;
        prev_->beginTuple();
        while (!prev_->is_end()) {
            auto record = prev_->Next();
            records.emplace_back(std::move(record));
            prev_->nextTuple();
            tuple_num++;
        }

        if (records.empty()) {
            return;
        }

        std::sort(
            records.begin(), records.end(),
            [this](const std::unique_ptr<RmRecord>& a,
                   const std::unique_ptr<RmRecord>& b) {
                char* a_val = new char[order_col.len];
                char* b_val = new char[order_col.len];
                memcpy(a_val, a->data + order_col.offset, order_col.len);
                memcpy(b_val, b->data + order_col.offset, order_col.len);
                bool res = false;
                switch (order_col.type) {
                    case TYPE_INT:
                        res = cmp_operator(*(int*)a_val, *(int*)b_val, OP_LT);
                        break;
                    case TYPE_FLOAT:
                        res =
                            cmp_operator(*(float*)a_val, *(float*)b_val, OP_LT);
                        break;
                    case TYPE_STRING:
                        res = cmp_operator(*(char*)a_val, *(char*)b_val, OP_LT);
                        break;
                }

                if (is_desc_) {
                    res = !res;
                }
                return res;
            });
    }

    void nextTuple() override { tuple_idx++; }

    std::unique_ptr<RmRecord> Next() override {
        return std::move(records[tuple_idx]);
    }

    bool is_end() const override { return tuple_idx == tuple_num; }

    Rid& rid() override { return _abstract_rid; }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    ColMeta get_col_offset(const TabCol& target) override {
        return *get_col(cols_, target);
    }
};