/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <string.h>
#include <cassert>
#include <cstdint>
#include <cstring>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>
#include "common/common.h"
#include "defs.h"
#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_seq_scan.h"
#include "index/ix.h"
#include "index/ix_index_handle.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class AggregationExecutor : public AbstractExecutor {

   private:
    struct AggregateValue {
        std::vector<Value> vals_;  // 聚合值
    };
    std::unique_ptr<AbstractExecutor> prev_;          // 子算子
    std::vector<ColMeta> cols_;                       // 所有输出的列
    std::vector<ColMeta> agg_cols_;                   // 需要聚合的列
    std::vector<AggrType> agg_types_;                 // 聚合类型
    std::vector<std::unique_ptr<RmRecord>> records_;  // 记录集
    AggregateValue agg;                               // 聚合值
    std::vector<ColMeta> group_cols_;                 // 分组列
    std::vector<Condition> having_conds_;             // having 条件
    std::map<std::string, int> group_map_;            // 分组列的映射
    std::vector<std::vector<std::unique_ptr<RmRecord>>>
        group_records_;  // 分组记录集
    // std::unordered_map<int, AggregateValue> agg_vals_map_;  // 聚合值映射

    bool end_ = false;
    bool is_group_by_ = false;
    bool is_empty_ = false;
    size_t len_;
    int group_idx_ = 0;
    int now_group_idx_ = 0;

    void val_init(ColMeta& col_meta, Value& val) {
        AggrType agg_type = col_meta.aggr_type;
        ColType col_type = col_meta.type;
        switch (col_type) {
            case TYPE_INT:
                val.type = TYPE_INT;
                break;

            case TYPE_FLOAT:
                val.type = TYPE_FLOAT;
                break;

            case TYPE_STRING:
                val.type = TYPE_STRING;
                break;
        }
        switch (agg_type) {
            case ::AggrType::AGG_SUM:
                val.init_zero();
                break;

            case ::AggrType::AGG_MAX:
                val.init_min(col_meta.len);
                break;

            case ::AggrType::AGG_MIN:
                val.init_max(col_meta.len);
                break;

            case ::AggrType::AGG_COUNT:
                val.type = TYPE_INT;
                val.init_zero();
                break;

            default:
                break;
        }
    }

   public:
    AggregationExecutor(std::unique_ptr<AbstractExecutor> prev,
                        std::vector<TabCol> sel_cols,
                        std::vector<TabCol> group_cols,
                        std::vector<Condition> having_conds) {
        prev_ = std::move(prev);
        len_ = 0;
        group_idx_ = 0, now_group_idx_ = 0;
        group_map_.clear();
        agg.vals_.clear();
        for (auto& col : sel_cols) {
            ColMeta col_meta;
            if (col.col_name == "*") {
                col_meta = ColMeta();
                col_meta.name = "*";
            } else {
                col_meta = prev_->get_col_offset(col);
            }
            agg_types_.emplace_back(col.aggr_type);
            col_meta.aggr_type = col.aggr_type;
            cols_.emplace_back(col_meta);

            if (col.aggr_type == ::AggrType::AGG_COUNT) {
                col_meta.len = sizeof(int);
                col_meta.type = TYPE_INT;
            }
            col_meta.offset = len_;
            agg_cols_.emplace_back(col_meta);

            len_ += col_meta.len;
        }

        // solve group_by and having conditions
        for (auto& col : group_cols) {
            auto col_meta = prev_->get_col_offset(col);
            group_cols_.emplace_back(col_meta);
        }
        is_group_by_ = group_cols_.size() > 0;
        having_conds_ = having_conds;
    }

    void init_agg_vals(AggregateValue& agg) {
        Value val;
        val.type = TYPE_INT, val.int_val = 0;
        for (size_t i = 0; i < agg_cols_.size(); i++) {
            agg.vals_.emplace_back(val);
        }
    }

    bool cmp_condition(const std::unique_ptr<RmRecord>& record,
                       const Condition& cond) {
        ColMeta lhs_cols_meta, rhs_cols_meta;
        Value agg_val;

        // cond.lhs_col is aggregate function
        if (cond.lhs_col.aggr_type != AGG_NULL) {
            if (cond.lhs_col.col_name == "*" &&
                cond.lhs_col.aggr_type == AGG_COUNT) {
                return cmp_operator<int>(records_.size(), cond.rhs_val.int_val,
                                         cond.op);
            } else {
                lhs_cols_meta = *get_col(cols_, cond.lhs_col);
                auto lhs_val = record->data + lhs_cols_meta.offset;
                agg_val =
                    generate_agg_val(lhs_cols_meta, cond.lhs_col.aggr_type);

                switch (agg_val.type) {
                    case TYPE_INT:
                        return cmp_operator<int>(agg_val.int_val,
                                                 cond.rhs_val.int_val, cond.op);
                    case TYPE_FLOAT:
                        return cmp_operator<float>(
                            agg_val.float_val, cond.rhs_val.float_val, cond.op);
                    case TYPE_STRING:
                        int str_len = 0;
                        while (str_len < lhs_cols_meta.len &&
                               lhs_val[str_len]) {
                            str_len++;
                        }
                        return cmp_operator<std::string>(
                            std::string(agg_val.str_val, str_len),
                            cond.rhs_val.str_val, cond.op);
                }
            }
        }

        lhs_cols_meta = *get_col(cols_, cond.lhs_col);

        char* rhs_val;
        ColType lhs_type, rhs_type;
        lhs_type = lhs_cols_meta.type;

        // cond.lhs_col is column value
        if (cond.is_rhs_val) {
            // 右边是val
            rhs_type = cond.rhs_val.type;
            rhs_val = cond.rhs_val.raw->data;
        } else {
            // 右边是col
            rhs_cols_meta = *get_col(cols_, cond.rhs_col);
            rhs_type = rhs_cols_meta.type;
            rhs_val = record->data + rhs_cols_meta.offset;
        }

        if (lhs_type != rhs_type) {
            throw InternalError("cmp_condition: Incompatible Comparision Type");
        }

        auto lhs_val = record->data + lhs_cols_meta.offset;
        switch (lhs_cols_meta.type) {
            case TYPE_INT:
                return cmp_operator<int>(*reinterpret_cast<int*>(lhs_val),
                                         *reinterpret_cast<int*>(rhs_val),
                                         cond.op);
            case TYPE_FLOAT:
                return cmp_operator<float>(*reinterpret_cast<float*>(lhs_val),
                                           *reinterpret_cast<float*>(rhs_val),
                                           cond.op);
            case TYPE_STRING:
                int str_len = 0;
                while (str_len < lhs_cols_meta.len && lhs_val[str_len]) {
                    str_len++;
                }
                return cmp_operator<std::string>(std::string(lhs_val, str_len),
                                                 cond.rhs_val.str_val, cond.op);
        }
        return false;
    }

    bool check_condition(const std::vector<Condition>& conds,
                         const std::unique_ptr<RmRecord>& record) {
        for (auto& cond : conds) {
            bool flag = cmp_condition(record, cond);
            if (!flag) {
                return false;
            }
        }
        return true;
    }

    Value generate_agg_val(ColMeta& col_meta, AggrType agg_type) {
        Value agg_val;
        if (col_meta.aggr_type != ::AggrType::AGG_NULL) {
            val_init(col_meta, agg_val);
        }

        switch (col_meta.aggr_type) {
            case ::AggrType::AGG_NULL: {
                break;
            }

            case ::AggrType::AGG_COUNT: {
                if (col_meta.name == "*") {
                    agg_val.int_val = records_.size();
                    break;
                }
                for (auto& record : records_) {
                    auto val = record->data + col_meta.offset;
                    if (col_meta.type == TYPE_STRING) {
                        std::string str_val(val, col_meta.len);
                        if (!str_val.empty()) {
                            agg_val.int_val++;
                        }
                    } else {
                        agg_val.int_val++;
                    }
                }
                break;
            }

            case ::AggrType::AGG_SUM: {
                for (auto& record : records_) {
                    auto val = record->data + col_meta.offset;
                    if (col_meta.type == TYPE_INT) {
                        agg_val.int_val += *(int*)val;
                    } else if (col_meta.type == TYPE_FLOAT) {
                        agg_val.float_val += *(float*)val;
                    } else if (col_meta.type == TYPE_STRING) {
                        throw AggregationError(
                            "sum is not support string type");
                    }
                }
                break;
            }

            case ::AggrType::AGG_MAX: {
                for (auto& record : records_) {
                    auto val = record->data + col_meta.offset;
                    if (col_meta.type == TYPE_INT) {
                        auto cmp_val = *(int*)val;
                        agg_val.int_val = std::max(agg_val.int_val, cmp_val);
                    } else if (col_meta.type == TYPE_FLOAT) {
                        auto cmp_val = *(float*)val;
                        agg_val.float_val =
                            std::max(agg_val.float_val, cmp_val);
                    } else if (col_meta.type == TYPE_STRING) {
                        throw AggregationError(
                            "max is not support string type");
                    }
                }
                break;
            }

            case ::AggrType::AGG_MIN: {
                for (auto& record : records_) {
                    auto val = record->data + col_meta.offset;
                    if (col_meta.type == TYPE_INT) {
                        auto cmp_val = *(int*)val;
                        agg_val.int_val = std::min(agg_val.int_val, cmp_val);
                    } else if (col_meta.type == TYPE_FLOAT) {
                        auto cmp_val = *(float*)val;
                        agg_val.float_val =
                            std::min(agg_val.float_val, cmp_val);
                    } else if (col_meta.type == TYPE_STRING) {
                        throw AggregationError(
                            "min is not support string type");
                    }
                }
                break;
            }
        }

        return agg_val;
    }

    bool precheck_records(const std::vector<Condition>& conds) {
        int idx = 0, remove_cnt = 0;
        std::vector<bool> removed_records(records_.size(), false);
        for (auto& record : records_) {
            if (!check_condition(conds, record)) {
                removed_records[idx] = true;
                remove_cnt++;
            }
            idx++;
        }
        if (remove_cnt == (int)records_.size()) {
            return false;
        }
        records_.erase(
            std::remove_if(records_.begin(), records_.end(),
                           [&removed_records,
                            this](const std::unique_ptr<RmRecord>& record) {
                               return removed_records[&record - &records_[0]];
                           }),
            records_.end());
        return true;
    }

    /**
     * @description: 
     */
    void beginTuple() override {
        prev_->beginTuple();
        while (!prev_->is_end()) {
            auto record = prev_->Next();

            std::string group_key = "";
            for (auto& col : group_cols_) {
                group_key += std::string(record->data + col.offset, col.len);
            }
            if (group_map_.find(group_key) == group_map_.end()) {
                group_map_[group_key] = group_idx_++;
                init_agg_vals(agg);
                group_records_.resize(group_idx_);
            }
            group_records_[group_map_[group_key]].emplace_back(
                std::move(record));

            prev_->nextTuple();
        }

        // 如果整个是空表，需要初始化一个空的group_records_，用于count(*)输出0
        if (group_records_.empty() &&
            std::any_of(
                agg_types_.begin(), agg_types_.end(),
                [](AggrType aggr_type) { return aggr_type == AGG_COUNT; })) {
            group_records_.resize(1);
            group_map_[""] = 0;
            group_idx_ = 1;
            init_agg_vals(agg);
        } else {
            nextTuple();
        }
    }

    /**
     * @description: 
     */
    void nextTuple() override {
        while (!is_end()) {
            if (now_group_idx_ >= (int)group_records_.size()) {
                end_ = true;
                break;
            }
            agg.vals_.clear();
            records_.clear();
            records_ = std::move(group_records_[now_group_idx_]);
            ++now_group_idx_;
            if (precheck_records(having_conds_)) {
                break;
            }
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        for (size_t i = 0; i < cols_.size(); i++) {
            auto& col_meta = cols_[i];
            auto& agg_val = agg.vals_[i];
            agg_val = generate_agg_val(col_meta, agg_types_[i]);
        }

        auto record = std::make_unique<RmRecord>(len_);
        for (size_t i = 0; i < agg_cols_.size(); i++) {
            auto& agg_col_meta = agg_cols_[i];
            if (agg_types_[i] == ::AggrType::AGG_NULL) {
                memcpy(record->data + agg_col_meta.offset,
                       records_[0]->data + cols_[i].offset, cols_[i].len);
                continue;
            }
            auto& agg_val = agg.vals_[i];
            switch (agg_col_meta.type) {
                case TYPE_INT: {
                    memcpy(record->data + agg_col_meta.offset, &agg_val.int_val,
                           agg_col_meta.len);
                    break;
                }

                case TYPE_FLOAT: {
                    memcpy(record->data + agg_col_meta.offset,
                           &agg_val.float_val, agg_col_meta.len);
                    break;
                }

                case TYPE_STRING: {
                    memcpy(record->data + agg_col_meta.offset, &agg_val,
                           agg_col_meta.len);
                    break;
                }
            }
        }

        return record;
    }

    Rid& rid() override { return _abstract_rid; }

    virtual std::string getType() override { return "AggregationExecutor"; }

    const std::vector<ColMeta>& cols() const override { return agg_cols_; };

    bool is_end() const override { return end_; }

    size_t tupleLen() const override { return len_; }
};
