/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cstring>
#include <string>
#include "common/common.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_seq_scan.h"
#include "index/ix.h"
#include "index/ix_scan.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表名称
    TabMeta tab_;                       // 表的元数据
    std::vector<Condition> conds_;      // 扫描条件
    RmFileHandle* fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // 需要读取的字段
    size_t len_;                        // 选取出来的一条记录的长度
    std::vector<Condition> fed_conds_;  // 扫描条件，和conds_字段相同
    int index_num_;                     // 有效索引数量

    std::vector<std::string>
        index_col_names_;   // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;  // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<RecScan> scan_;

    SmManager* sm_manager_;

   public:
    IndexScanExecutor(SmManager* sm_manager, std::string tab_name,
                      std::vector<Condition> conds,
                      std::vector<std::string> index_col_names,
                      Context* context) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        conds_ = std::move(conds);
        // index_no_ = index_no;
        index_col_names_ = index_col_names;
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT},
            {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        for (auto& cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                // lhs is on other table, now rhs must be on this table
                assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
                // swap lhs and rhs
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
        }
        // fed_conds_ = conds_;

        for (int i = 0; i < index_meta_.col_num; i++) {
            bool found = false;
            for (auto& cond : conds_) {
                if (cond.lhs_col.col_name == index_meta_.cols[i].name) {
                    found = true;
                    fed_conds_.push_back(cond);
                }
            }
            if (!found) {
                break;
            }
            index_num_++;
        }

        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
    }

    bool cmp_condition(const std::unique_ptr<RmRecord>& record,
                       const Condition& cond) {
        auto lhs_cols_meta = get_col(cols_, cond.lhs_col);
        auto lhs_val = record->data + lhs_cols_meta->offset;
        char* rhs_val;
        ColType lhs_type, rhs_type;
        lhs_type = lhs_cols_meta->type;
        if (cond.is_rhs_val) {
            // 右边是val
            rhs_type = cond.rhs_val.type;
            rhs_val = cond.rhs_val.raw->data;
        } else {
            // 右边是col
            auto rhs_cols_meta = get_col(cols_, cond.rhs_col);
            rhs_type = rhs_cols_meta->type;
            rhs_val = record->data + rhs_cols_meta->offset;
        }

        if (lhs_type != rhs_type) {
            throw InternalError("cmp_condition: Incompatible Comparision Type");
        }

        switch (lhs_cols_meta->type) {
            case TYPE_INT:
                return cmp_operator<int>(*reinterpret_cast<int*>(lhs_val),
                                         *reinterpret_cast<int*>(rhs_val),
                                         cond.op);
            case TYPE_FLOAT:
                return cmp_operator<float>(*reinterpret_cast<float*>(lhs_val),
                                           *reinterpret_cast<float*>(rhs_val),
                                           cond.op);
            case TYPE_STRING:
                int str_len = 0;
                while (str_len < lhs_cols_meta->len && lhs_val[str_len]) {
                    str_len++;
                }
                return cmp_operator<std::string>(std::string(lhs_val, str_len),
                                                 cond.rhs_val.str_val, cond.op);
        }
        return false;
    }

    bool check_condition(const std::unique_ptr<RmRecord>& record) {
        for (auto& cond : fed_conds_) {
            bool flag = cmp_condition(record, cond);
            if (!flag) {
                return false;
            }
        }
        return true;
    }

    void beginTuple() override {

        auto index_name = sm_manager_->get_ix_manager()->get_index_name(
            tab_name_, index_col_names_);

        for (auto& [x, y] : sm_manager_->ihs_) {
            spdlog::info("index name:{}, file ptr:{}", x, (char*)y.get());
        }
        auto ih = sm_manager_->ihs_[index_name].get();

        // 这里就是分类讨论，根据各种不同的 >= > = < <= 的条件，来确定lower和upper bound
        // 然后用lower_bound和upper_bound来初始化IxScan，从而达到走索引去查询的效果
        auto lower = ih->leaf_begin(), upper = ih->leaf_end();

        char* upper_key = new char[index_meta_.col_tot_len];
        char* lower_key = new char[index_meta_.col_tot_len];
        // printf("col len = %d\n", index_meta_.col_tot_len);

        int key_offset = 0;
        for (int i = 0; i < index_meta_.col_num; i++) {
            auto col_meta = index_meta_.cols[i];
            // 初始化，先默认取最大值和最小值作为本索引列的lower_key和upper_key
            Value max_val, min_val;
            max_val.type = col_meta.type;
            min_val.type = col_meta.type;
            max_val.init_max(col_meta.len);
            min_val.init_min(col_meta.len);
            memcpy(lower_key + key_offset, min_val.raw->data, col_meta.len);
            memcpy(upper_key + key_offset, max_val.raw->data, col_meta.len);

            if (i < index_num_) {
                // 对于在索引列上的条件，根据条件的类型，设置lower_key和upper_key
                for (auto& cond : fed_conds_) {
                    if (cond.lhs_col.col_name != col_meta.name) {
                        continue;
                    }
                    if (cond.op == OP_NE) {
                    } else if (cond.op == OP_EQ) {
                        memcpy(lower_key + key_offset, cond.rhs_val.raw->data,
                               col_meta.len);
                        memcpy(upper_key + key_offset, cond.rhs_val.raw->data,
                               col_meta.len);
                    } else if (cond.op == OP_GE || cond.op == OP_GT) {
                        memcpy(lower_key + key_offset, cond.rhs_val.raw->data,
                               col_meta.len);
                    } else if (cond.op == OP_LE || cond.op == OP_LT) {
                        memcpy(upper_key + key_offset, cond.rhs_val.raw->data,
                               col_meta.len);
                    }
                }
            }
            key_offset += col_meta.len;
        }

        lower = ih->lower_bound(lower_key);
        upper = ih->upper_bound(upper_key);

        delete[] lower_key;
        delete[] upper_key;

        scan_ =
            std::make_unique<IxScan>(ih, lower, upper, sm_manager_->get_bpm());
        while (!scan_->is_end()) {
            rid_ = scan_->rid();
            auto record = fh_->get_record(rid_, context_);
            if (check_condition(record)) {
                break;
            }
            scan_->next();
        }
    }

    void nextTuple() override {
        if (!scan_->is_end()) {
            scan_->next();
        }

        while (!scan_->is_end()) {
            rid_ = scan_->rid();
            auto record = fh_->get_record(rid_, context_);
            if (check_condition(record)) {
                break;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if (!fh_->is_record(rid_)) {
            return nullptr;
        }

        auto record = fh_->get_record(rid_, context_);
        return record;
    }

    bool is_end() const override { return scan_->is_end(); }

    size_t tupleLen() const override { return len_; }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    Rid& rid() override { return rid_; }
};
