/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include "transaction/txn_defs.h"

class InsertExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                // 表的元数据
    std::vector<Value> values_;  // 需要插入的数据
    RmFileHandle* fh_;           // 表的数据文件句柄
    std::string tab_name_;       // 表名称
    Rid rid_;  // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
    SmManager* sm_manager_;

   public:
    InsertExecutor(SmManager* sm_manager, const std::string& tab_name,
                   std::vector<Value> values, Context* context) {
        sm_manager_ = sm_manager;
        tab_ = sm_manager_->db_.get_table(tab_name);
        values_ = values;
        tab_name_ = tab_name;
        if (values.size() != tab_.cols.size()) {
            throw InvalidValueCountError();
        }
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        context_ = context;
        context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd());
        // context_->lock_mgr_->lock_exclusive_on_table(context_->txn_,
        //                                              fh_->GetFd());
    };

    std::unique_ptr<RmRecord> Next() override {
        // Make record buffer
        RmRecord rec(fh_->get_file_hdr().record_size);
        for (size_t i = 0; i < values_.size(); i++) {
            auto& col = tab_.cols[i];
            auto& val = values_[i];
            val.type_convert_init(col);
            if (col.type != val.type) {
                throw IncompatibleTypeError(coltype2str(col.type),
                                            coltype2str(val.type));
            }
            // val.init_raw(col.len);
            memcpy(rec.data + col.offset, val.raw->data, col.len);
        }

        std::vector<std::unique_ptr<RmRecord>> records;

        // Insert into index
        for (size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_
                          .at(sm_manager_->get_ix_manager()->get_index_name(
                              tab_name_, index.cols))
                          .get();
            auto key = index.get_index_key(rec.data);

            // Check if the key already exists in the index
            std::vector<Rid> res;
            if (ih->get_value(key.get(), &res, context_->txn_)) {
                throw IndexNotUniqueError();
            }

            auto cur_rec =
                std::make_unique<RmRecord>(index.col_tot_len, key.get());
            records.push_back(std::move(cur_rec));
            // ih->insert_entry(key, rid_, context_->txn_);
        }

        // Insert into record file
        // 先做这个，再做索引，因为要更新rid_
        rid_ = fh_->insert_record(rec.data, context_);

        // Insert into index
        for (size_t i = 0; i < records.size(); i++) {
            auto& record = records[i];
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_
                          .at(sm_manager_->get_ix_manager()->get_index_name(
                              tab_name_, index.cols))
                          .get();
            ih->insert_entry(record->data, rid_, context_->txn_);
        }

        if (context_->txn_->get_txn_mode()) {
            WriteRecord* write_record =
                new WriteRecord(WType::INSERT_TUPLE, tab_name_, rid_, rec);
            context_->txn_->append_write_record(write_record);
        }

        return nullptr;
    }
    Rid& rid() override { return rid_; }

    virtual std::string getType() override { return "InsertExecutor"; }
};
