/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>
#include <cstring>
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_seq_scan.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;  // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;  // 右儿子节点（需要join的表）
    size_t len_;                 // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;  // join后获得的记录的字段

    std::vector<Condition> fed_conds_;  // join条件
    bool isend;

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left,
                           std::unique_ptr<AbstractExecutor> right,
                           std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto& col : right_cols) {
            col.offset += left_->tupleLen();
        }

        // cols_ 包括左右两表的所有字段
        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);
    }

    bool cmp_condition(const std::unique_ptr<RmRecord>& left_record,
                       const std::unique_ptr<RmRecord>& right_record,
                       const Condition& cond) {
        auto lhs_cols_meta = get_col(cols_, cond.lhs_col);
        auto lhs_val = left_record->data + lhs_cols_meta->offset;
        auto lhs_type = lhs_cols_meta->type;

        // 右边是col
        auto rhs_cols_meta = get_col(cols_, cond.rhs_col);
        // 现在cols_是左右两表的所有字段，所以在使用右表的offset时需要减去左表的长度
        auto rhs_val =
            right_record->data + rhs_cols_meta->offset - left_record->size;
        auto rhs_type = rhs_cols_meta->type;

        if (lhs_type != rhs_type) {
            throw InternalError("cmp_condition: Incompatible Comparision Type");
        }

        switch (lhs_cols_meta->type) {
            case TYPE_INT:
                return cmp_operator<int>(*reinterpret_cast<int*>(lhs_val),
                                         *reinterpret_cast<int*>(rhs_val),
                                         cond.op);
            case TYPE_FLOAT:
                return cmp_operator<float>(*reinterpret_cast<float*>(lhs_val),
                                           *reinterpret_cast<float*>(rhs_val),
                                           cond.op);
            case TYPE_STRING:
                int str_len = 0;
                while (str_len < lhs_cols_meta->len && lhs_val[str_len]) {
                    str_len++;
                }
                return cmp_operator<std::string>(std::string(lhs_val, str_len),
                                                 cond.rhs_val.str_val, cond.op);
        }
        return false;
    }

    bool check_condition(const std::unique_ptr<RmRecord>& left_record,
                         const std::unique_ptr<RmRecord>& right_record) {
        for (auto& cond : fed_conds_) {
            bool flag = cmp_condition(left_record, right_record, cond);
            if (!flag) {
                return false;
            }
        }
        return true;
    }

    void beginTuple() override {
        left_->beginTuple();
        right_->beginTuple();

        next_valid_tuple();
    }

    void nextTuple() override {
        if (isend) {
            return;
        }

        if (right_->is_end()) {
            left_->nextTuple();
            right_->beginTuple();
        } else {
            right_->nextTuple();
        }

        next_valid_tuple();
    }

    void next_valid_tuple() {
        while (!left_->is_end()) {
            auto left_record = left_->Next();

            while (!right_->is_end()) {
                auto right_record = right_->Next();
                if (check_condition(left_record, right_record)) {
                    return;
                }
                right_->nextTuple();
            }
            left_->nextTuple();
            right_->beginTuple();
        }
        isend = true;
    }

    std::unique_ptr<RmRecord> Next() override {
        if (is_end()) {
            return nullptr;
        }

        auto left_record = left_->Next();
        auto right_record = right_->Next();
        auto record = std::make_unique<RmRecord>(len_);

        memcpy(record->data, left_record->data, left_record->size);
        memcpy(record->data + left_record->size, right_record->data,
               right_record->size);
        return record;
    }

    Rid& rid() override { return _abstract_rid; }

    const std::vector<ColMeta>& cols() const override { return cols_; };

    bool is_end() const override { return isend; }

    size_t tupleLen() const override { return len_; }
};