/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class ProjectionExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;  // 投影节点的儿子节点
    std::vector<ColMeta> cols_;               // 需要投影的字段
    size_t len_;                              // 字段总长度
    bool has_aggr_ = false;                   // 是否有聚合函数
    std::vector<size_t> sel_idxs_;  // 选中的字段在表中是第几个

   public:
    ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev,
                       const std::vector<TabCol>& sel_cols, bool has_aggr) {
        prev_ = std::move(prev);
        has_aggr_ = has_aggr;
        if (!has_aggr_) {
            size_t curr_offset = 0;
            auto& prev_cols = prev_->cols();
            for (auto& sel_col : sel_cols) {
                auto pos = get_col(prev_cols, sel_col);
                sel_idxs_.push_back(pos - prev_cols.begin());
                auto col = *pos;
                col.offset = curr_offset;
                curr_offset += col.len;
                cols_.push_back(col);
            }

            len_ = curr_offset;
        }
    }

    void beginTuple() override { prev_->beginTuple(); }

    void nextTuple() override { prev_->nextTuple(); }

    std::unique_ptr<RmRecord> Next() override {
        if (has_aggr_) {
            return prev_->Next();
        }
        auto prev_record = prev_->Next();  // 得到满足儿子节点条件的记录
        auto& prev_cols = prev_->cols();
        RmRecord record(len_);  // 投影的记录

        for (size_t i = 0; i < cols_.size(); i++) {
            const auto& offset_prev = sel_idxs_[i];

            // sel_idxs_[i] 记录的是在 prev_cols 中的位置
            // cols_[i].offset 记录的是在 record 中的位置
            // cols_[i].len 记录的是需要复制的长度
            // 现在要把 prev_record->data + offset_prev 复制到 record->data + cols_[i].offset
            memcpy(record.data + cols_[i].offset,
                   prev_record->data + prev_cols[offset_prev].offset,
                   cols_[i].len);
        }
        return std::make_unique<RmRecord>(std::move(record));
    }

    Rid& rid() override { return _abstract_rid; }

    const std::vector<ColMeta>& cols() const override {
        return has_aggr_ ? prev_->cols() : cols_;
    };

    bool is_end() const override { return prev_->is_end(); }

    size_t tupleLen() const override { return len_; }
};
