/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <string.h>
#include <cstdint>
#include <cstring>
#include <memory>
#include <vector>
#include "defs.h"
#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "index/ix_index_handle.h"
#include "system/sm.h"

template <typename T>
inline bool cmp_operator(const T& a, const T& b, CompOp op) {
    switch (op) {
        case OP_EQ:
            return a == b;
        case OP_NE:
            return a != b;
        case OP_LT:
            return a < b;
        case OP_GT:
            return a > b;
        case OP_LE:
            return a <= b;
        case OP_GE:
            return a >= b;
        default:
            throw InternalError("cmp_operator: Unsupported OP Type");
    }
    return false;
}

// template <>
// inline bool cmp_operator<const char*>(const char* const& a,
//                                       const char* const& b, CompOp op, int len) {
//     auto cmp_ret = memcmp(a, b, len);
//     switch (op) {
//         case OP_EQ:
//             return cmp_ret == 0;
//         case OP_NE:
//             return cmp_ret != 0;
//         case OP_LT:
//             return cmp_ret < 0;
//         case OP_GT:
//             return cmp_ret > 0;
//         case OP_LE:
//             return cmp_ret <= 0;
//         case OP_GE:
//             return cmp_ret >= 0;
//         default:
//             throw InternalError("cmp_operator: Unsupported OP Type");
//     }
//     return false;
// }

class SeqScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    RmFileHandle* fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

    Rid rid_;
    std::unique_ptr<RecScan> scan_;  // table_iterator

    SmManager* sm_manager_;

   public:
    SeqScanExecutor(SmManager* sm_manager, std::string tab_name,
                    std::vector<Condition> conds, Context* context) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        TabMeta& tab = sm_manager_->db_.get_table(tab_name_);
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;

        context_ = context;

        fed_conds_ = conds_;
        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
    }

    bool cmp_condition(const std::unique_ptr<RmRecord>& record,
                       const Condition& cond) {
        auto lhs_cols_meta = get_col(cols_, cond.lhs_col);
        auto lhs_val = record->data + lhs_cols_meta->offset;
        char* rhs_val;
        ColType lhs_type, rhs_type;
        lhs_type = lhs_cols_meta->type;
        if (cond.is_rhs_val) {
            // 右边是val
            rhs_type = cond.rhs_val.type;
            rhs_val = cond.rhs_val.raw->data;
        } else {
            // 右边是col
            auto rhs_cols_meta = get_col(cols_, cond.rhs_col);
            rhs_type = rhs_cols_meta->type;
            rhs_val = record->data + rhs_cols_meta->offset;
        }

        if (lhs_type != rhs_type) {
            throw InternalError("cmp_condition: Incompatible Comparision Type");
        }

        switch (lhs_cols_meta->type) {
            case TYPE_INT:
                return cmp_operator<int>(*reinterpret_cast<int*>(lhs_val),
                                         *reinterpret_cast<int*>(rhs_val),
                                         cond.op);
            case TYPE_FLOAT:
                return cmp_operator<float>(*reinterpret_cast<float*>(lhs_val),
                                           *reinterpret_cast<float*>(rhs_val),
                                           cond.op);
            case TYPE_STRING:
                int str_len = 0;
                while (str_len < lhs_cols_meta->len && lhs_val[str_len]) {
                    str_len++;
                }
                return cmp_operator<std::string>(std::string(lhs_val, str_len),
                                                 cond.rhs_val.str_val, cond.op);
        }
        return false;
    }

    bool check_condition(const std::unique_ptr<RmRecord>& record) {
        for (auto& cond : conds_) {
            bool flag = cmp_condition(record, cond);
            if (!flag) {
                return false;
            }
        }
        return true;
    }

    /**
     * @description: 初始化scan_，并让scan_扫描到第一个满足条件的记录再停止，即调一下nextTuple()就行
     */
    void beginTuple() override {
        scan_ = std::make_unique<RmScan>(fh_);
        while (!scan_->is_end()) {
            // auto rid = scan_->rid();
            rid_ = scan_->rid();
            auto record = fh_->get_record(rid_, context_);
            if (check_condition(record)) {
                // rid_ = rid;
                break;
            }
            scan_->next();
        }
    }

    /**
     * @description: 从当前scan_对象中获取下一条满足条件的记录，并更新 rid_
     */
    void nextTuple() override {
        if (!scan_->is_end()) {
            scan_->next();
        }

        while (!scan_->is_end()) {
            // auto rid = scan_->rid();
            rid_ = scan_->rid();
            auto record = fh_->get_record(rid_, context_);
            if (check_condition(record)) {
                // rid_ = rid;
                break;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        /** STEP:
            1. check if the rid is valid
            2. get the record from the file handle
            3. check if the record is valid with the conditions
                3.1 根据 planer.cpp::pop_conds 得到的条件，对 record 进行过滤, 条件包括：{对应表字段 op 值}, {同表字段 op 同表字段}
                3.2 如果条件不满足，则跳过当前记录，并调用 nextTuple() 继续获取下一条记录
            4. return the record
        */
        if (!fh_->is_record(rid_)) {
            // 对应的rid没有记录
            // check: throw exception?
            return nullptr;
            // throw RecordNotFoundError(rid_.page_no, rid_.slot_no);
        }

        auto record = fh_->get_record(rid_, context_);
        return record;
    }

    Rid& rid() override { return rid_; }

    virtual std::string getType() override { return "SeqScanExecutor"; }

    const std::vector<ColMeta>& cols() const override { return cols_; };

    bool is_end() const override { return scan_->is_end(); }

    size_t tupleLen() const override { return len_; }

    ColMeta get_col_offset(const TabCol& target) override {
        for (auto col : cols_) {
            if (col.name == target.col_name &&
                col.tab_name == target.tab_name) {
                return col;
            }
        }
        throw ColumnNotFoundError(target.tab_name + "." + target.col_name);
    };
};
