/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstring>
#include <memory>
#include <vector>
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "system/sm.h"
#include "system/sm_meta.h"
#include "transaction/txn_defs.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle* fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    std::vector<ColMeta> cols_;
    SmManager* sm_manager_;

    void update_record(const std::unique_ptr<RmRecord>& record, Rid rid) {
        std::vector<std::unique_ptr<RmRecord>> old_keys;
        std::vector<std::unique_ptr<RmRecord>> new_keys;

        for (const auto& set_clause : set_clauses_) {
            // set_clause.lhs is column, set_clause.rhs is value.
            // TODO: maybe should support set_clause.rhs is expression or column.
            auto tab_col = set_clause.lhs;
            auto val = set_clause.rhs;
            auto tab_name = tab_.name;
            auto col_name = tab_col.col_name;
            auto set_col_meta = get_col(cols_, tab_col);
            val.type_convert_init(*set_col_meta);
            memcpy(record->data + set_col_meta->offset, val.raw->data,
                   set_col_meta->len);
        }
    }

   public:
    UpdateExecutor(SmManager* sm_manager, const std::string& tab_name,
                   std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids,
                   Context* context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        cols_ = tab_.cols;
        context_ = context;
        context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd());
    }
    std::unique_ptr<RmRecord> Next() override {
        for (auto& rid : rids_) {
            auto record = fh_->get_record(rid, context_);
            if (record == nullptr) {
                continue;
            }
            auto upd_record = std::make_unique<RmRecord>(record->size);
            memcpy(upd_record->data, record->data, record->size);

            for (const auto& set_clause : set_clauses_) {
                // set_clause.lhs is column, set_clause.rhs is value.
                // TODO: maybe should support set_clause.rhs is expression or column.
                auto tab_col = set_clause.lhs;
                auto val = set_clause.rhs;
                auto tab_name = tab_.name;
                auto col_name = tab_col.col_name;
                auto set_col_meta = get_col(cols_, tab_col);
                val.type_convert_init(*set_col_meta);
                memcpy(upd_record->data + set_col_meta->offset, val.raw->data,
                       set_col_meta->len);
            }

            for (size_t i = 0; i < tab_.indexes.size(); i++) {
                auto& index = tab_.indexes[i];
                auto ih = sm_manager_->ihs_
                              .at(sm_manager_->get_ix_manager()->get_index_name(
                                  tab_name_, index.cols))
                              .get();

                auto old_key = index.get_index_key(record->data);
                auto new_key = index.get_index_key(upd_record->data);

                if (!memcmp(new_key.get(), old_key.get(), index.col_tot_len)) {
                    continue;
                }

                std::vector<Rid> res;
                if (ih->get_value(new_key.get(), &res, context_->txn_)) {
                    throw IndexNotUniqueError();
                }
                ih->delete_entry(old_key.get(), context_->txn_);
                ih->insert_entry(new_key.get(), rid, context_->txn_);
            }

            if (context_->txn_->get_txn_mode()) {
                RmRecord rec{upd_record->size};
                // TODO: can we use std::move here to avoid copying? but update_record() will use the upd_record.
                memcpy(rec.data, upd_record->data, upd_record->size);
                WriteRecord* write_record =
                    new WriteRecord(WType::UPDATE_TUPLE, tab_name_, rid, rec);
                write_record->old_record_ = *std::move(record);
                context_->txn_->append_write_record(write_record);
            }

            fh_->update_record(rid, upd_record->data, context_);
        }
        return nullptr;
    }

    Rid& rid() override { return _abstract_rid; }
};