#include <memory>
#include <vector>
#include "common/common.h"
#include "defs.h"
#include "gtest/gtest.h"
#include "index/ix_manager.h"
#include "record/rm.h"
#include "record/rm_defs.h"
#include "record/rm_manager.h"
#include "recovery/log_recovery.h"
#include "storage/buffer_pool_manager.h"
#include "system/sm_manager.h"
#include "transaction/concurrency/lock_manager.h"
#include "transaction/transaction_manager.h"

#define private public
#include "index/ix.h"
#undef private  // for use private variables in "ix.h"

const std::string TEST_DB_NAME = "LogTestDB";
const std::string TEST_FILE_NAME = "table1";         // 测试文件名前缀
const std::vector<std::string> TEST_COL = {"col1"};  // 索引列

class LogManagerTest : public ::testing::Test {
   public:
    std::unique_ptr<LockManager> lock_manager_;
    std::unique_ptr<DiskManager> disk_manager_;
    std::unique_ptr<BufferPoolManager> buffer_pool_manager_;
    std::unique_ptr<RmManager> rm_manager_;
    std::unique_ptr<IxManager> ix_manager_;
    std::unique_ptr<SmManager> sm_manager_;
    std::unique_ptr<LogManager> log_manager_;
    std::unique_ptr<TransactionManager> txn_manager_;
    std::unique_ptr<Transaction> txn_;
    std::unique_ptr<RecoveryManager> recovery_;
    IxIndexHandle* ih_;

   public:
    void SetUp() override {
        ::testing::Test::SetUp();
        lock_manager_ = std::make_unique<LockManager>();
        disk_manager_ = std::make_unique<DiskManager>();
        buffer_pool_manager_ = std::make_unique<BufferPoolManager>(
            BUFFER_POOL_SIZE, disk_manager_.get());
        rm_manager_ = std::make_unique<RmManager>(disk_manager_.get(),
                                                  buffer_pool_manager_.get());
        ix_manager_ = std::make_unique<IxManager>(disk_manager_.get(),
                                                  buffer_pool_manager_.get());
        txn_ = std::make_unique<Transaction>(0);
        sm_manager_ = std::make_unique<SmManager>(
            disk_manager_.get(), buffer_pool_manager_.get(), rm_manager_.get(),
            ix_manager_.get());
        log_manager_ = std::make_unique<LogManager>(disk_manager_.get());
        txn_manager_ = std::make_unique<TransactionManager>(lock_manager_.get(),
                                                            sm_manager_.get());
        recovery_ = std::make_unique<RecoveryManager>(
            disk_manager_.get(), buffer_pool_manager_.get(), sm_manager_.get());
    }

    // 清理函数，每个测试用例运行结束后调用
    void TearDown() override {
        // 清理管理器实例
        lock_manager_.reset();
        disk_manager_.reset();
        buffer_pool_manager_.reset();
        rm_manager_.reset();
        ix_manager_.reset();
        sm_manager_.reset();
        log_manager_.reset();
        txn_manager_.reset();
        txn_.reset();
        recovery_.reset();
        ih_ = nullptr;
        ::testing::Test::TearDown();
    }

    void init_txn_and_rows(std::vector<Transaction*>& txns,
                           std::vector<Rid>& rows, int num) {
        for (int i = 0; i < num; ++i) {
            rows.push_back({0, i});
            Transaction* t = new Transaction(i);
            txns.emplace_back(t);
            EXPECT_EQ(i, txns[i]->get_transaction_id());  // 验证事务ID
        }
    }

    void clear_txns(std::vector<Transaction*>& txns) {
        for (Transaction* txn : txns) {
            delete txn;
        }
        txns.clear();
    }

    void commit_txn(Transaction* txn) {
        txn_manager_->commit(txn, log_manager_.get());
        EXPECT_EQ(txn->get_state(), TransactionState::COMMITTED);
    }

    void abort_txn(Transaction* txn) {
        txn_manager_->abort(txn, log_manager_.get());
        EXPECT_EQ(txn->get_state(), TransactionState::ABORTED);
    }

    std::vector<RmRecord> get_insert_record(TabMeta& tab_meta) {
        std::vector<Value> vals;
        std::vector<RmRecord> recs;
        auto fh = sm_manager_->fhs_[TEST_FILE_NAME].get();
        RmRecord rec(fh->get_file_hdr().record_size);
        for (int j = 0; j < 20; j++) {
            for (size_t i = 0; i < tab_meta.cols.size(); i++) {
                auto& col = tab_meta.cols[i];
                if (i == 0) {
                    Value x;
                    x.set_int(j);
                    x.init_raw(4);
                    vals.push_back(x);
                } else if (i == 1) {
                    Value x;
                    x.set_str("TestRecord");
                    x.init_raw(10);
                    vals.push_back(x);
                }

                memcpy(rec.data + col.offset, vals[i].raw->data, col.len);
            }
            vals.clear();
            recs.push_back(rec);
        }
        return recs;
    }
};

TEST_F(LogManagerTest, BasicTest1_CloseDB_correct) {
    // 初始化事务和记录
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 3;
    init_txn_and_rows(txns, rows_, num);

    // 建DB
    if (disk_manager_->is_dir(TEST_DB_NAME)) {
        std::string cmd = "rm -rf " + TEST_DB_NAME;
        if (system(cmd.c_str()) < 0) {
            throw UnixError();
        }
    }
    sm_manager_->create_db(TEST_DB_NAME);
    sm_manager_->open_db(TEST_DB_NAME);
    auto& db_ = sm_manager_->db_;

    if (ix_manager_->exists(TEST_FILE_NAME, TEST_COL)) {
        ix_manager_->destroy_index(TEST_FILE_NAME, TEST_COL);
    }

    // 建表
    std::vector<ColDef> coldef;
    coldef.push_back({"col1", TYPE_INT, 4});
    coldef.push_back({"col2", TYPE_STRING, 10});

    auto cur_context = std::make_shared<Context>(
        lock_manager_.get(), log_manager_.get(), txn_.get());

    sm_manager_->create_table(TEST_FILE_NAME, coldef, cur_context.get());
    sm_manager_->create_index(TEST_FILE_NAME, TEST_COL, cur_context.get());
    assert(ix_manager_->exists(TEST_FILE_NAME, TEST_COL));

    std::string ix_name = ix_manager_->get_index_name(TEST_FILE_NAME, TEST_COL);
    ih_ = sm_manager_->ihs_[ix_name].get();

    auto fh = sm_manager_->fhs_[TEST_FILE_NAME].get();
    auto& tab_meta = db_.get_table(TEST_FILE_NAME);
    // 插入记录并提交事务

    auto recs = get_insert_record(tab_meta);
    std::vector<Rid> rids;

    // 事务1：插入并提交
    cur_context->txn_ = txns[0];
    auto begin_log_record = BeginLogRecord();
    begin_log_record.log_tid_ = txns[0]->get_transaction_id();
    begin_log_record.lsn_ = 0;
    log_manager_->add_log_to_buffer(&begin_log_record);
    for (size_t j = 0; j < recs.size() / 2; j++) {
        auto rid = fh->insert_record(recs[j].data, cur_context.get());
        rids.push_back(rid);
    }
    commit_txn(txns[0]);

    // 事务2：插入但不提交（模拟系统崩溃前的操作）
    cur_context->txn_ = txns[1];
    begin_log_record = BeginLogRecord();
    begin_log_record.log_tid_ = txns[0]->get_transaction_id();
    begin_log_record.lsn_ = 0;
    for (size_t j = recs.size() / 2; j < recs.size(); j++) {
        auto rid = fh->insert_record(recs[j].data, cur_context.get());
        rows_.push_back(rid);
    }
    // 没有commit，模拟系统崩溃

    // log 刷盘
    log_manager_->flush_log_to_disk();

    sm_manager_->close_db();
    // 关闭数据库，模拟系统关闭
    TearDown();

    // 重新初始化系统，模拟系统重启
    SetUp();
    sm_manager_->open_db(TEST_DB_NAME);
    fh = sm_manager_->fhs_[TEST_FILE_NAME].get();
    auto new_context = std::make_shared<Context>(
        lock_manager_.get(), log_manager_.get(), txn_.get());

    // 尝试获取刚才插入的记录
    for (size_t j = 0; j < rids.size(); j++) {
        auto fetched_record = fh->get_record(rids[j], new_context.get());

        // 检查记录是否与插入的记录一致
        ASSERT_EQ(fetched_record->size, recs[j].size);
        ASSERT_EQ(
            memcmp(fetched_record->data, recs[j].data, fetched_record->size),
            0);
    }
}

TEST_F(LogManagerTest, BasicTest2_CloseDB_crash) {
    // 初始化事务和记录
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 3;
    init_txn_and_rows(txns, rows_, num);

    // 建DB
    if (disk_manager_->is_dir(TEST_DB_NAME)) {
        std::string cmd = "rm -rf " + TEST_DB_NAME;
        if (system(cmd.c_str()) < 0) {
            throw UnixError();
        }
    }
    sm_manager_->create_db(TEST_DB_NAME);
    sm_manager_->open_db(TEST_DB_NAME);
    recovery_->analyze();
    recovery_->redo();
    recovery_->undo();

    if (ix_manager_->exists(TEST_FILE_NAME, TEST_COL)) {
        ix_manager_->destroy_index(TEST_FILE_NAME, TEST_COL);
    }

    // 建表
    std::vector<ColDef> coldef;
    coldef.push_back({"col1", TYPE_INT, 4});
    coldef.push_back({"col2", TYPE_STRING, 10});

    auto cur_context = std::make_shared<Context>(
        lock_manager_.get(), log_manager_.get(), txn_.get());

    sm_manager_->create_table(TEST_FILE_NAME, coldef, cur_context.get());
    sm_manager_->create_index(TEST_FILE_NAME, TEST_COL, cur_context.get());
    assert(ix_manager_->exists(TEST_FILE_NAME, TEST_COL));

    std::string ix_name = ix_manager_->get_index_name(TEST_FILE_NAME, TEST_COL);
    ih_ = sm_manager_->ihs_[ix_name].get();

    auto& db_ = sm_manager_->db_;
    auto fh = sm_manager_->fhs_[TEST_FILE_NAME].get();
    auto& tab_meta = db_.get_table(TEST_FILE_NAME);
    // 插入记录并提交事务

    auto recs = get_insert_record(tab_meta);
    std::vector<Rid> rids;

    // 事务1：插入并提交
    cur_context->txn_ = txns[0];
    auto begin_log_record = BeginLogRecord();
    begin_log_record.log_tid_ = txns[0]->get_transaction_id();
    begin_log_record.lsn_ = 0;
    log_manager_->add_log_to_buffer(&begin_log_record);
    for (size_t j = 0; j < recs.size() / 2; j++) {
        auto rid = fh->insert_record(recs[j].data, cur_context.get());
        rids.push_back(rid);
    }
    commit_txn(txns[0]);

    // 事务2：插入但不提交（模拟系统崩溃前的操作）
    cur_context->txn_ = txns[1];
    begin_log_record = BeginLogRecord();
    begin_log_record.log_tid_ = txns[1]->get_transaction_id();
    begin_log_record.lsn_ = 12, begin_log_record.prev_lsn_ = 11;
    log_manager_->add_log_to_buffer(&begin_log_record);
    for (size_t j = recs.size() / 2; j < recs.size(); j++) {
        auto rid = fh->insert_record(recs[j].data, cur_context.get());
        rids.push_back(rid);
    }
    // 没有commit，模拟系统崩溃

    // log 刷盘
    log_manager_->flush_log_to_disk();

    // sm_manager_->close_db();
    // 不关闭数据库，模拟系统 crash
    TearDown();

    // 重新初始化系统，模拟系统重启
    SetUp();

    if (chdir("..") < 0) {
        throw UnixError();
    }
    sm_manager_->open_db(TEST_DB_NAME);

    // !! add Recover->Redo here
    recovery_->analyze();
    recovery_->redo();
    recovery_->undo();

    fh = sm_manager_->fhs_[TEST_FILE_NAME].get();
    auto new_context = std::make_shared<Context>(
        lock_manager_.get(), log_manager_.get(), txn_.get());

    // 尝试获取刚才插入的记录
    // 事务1：应存在
    for (size_t j = 0; j < rids.size() / 2; j++) {
        auto fetched_record = fh->get_record(rids[j], new_context.get());

        // 检查记录是否与插入的记录一致
        ASSERT_EQ(fetched_record->size, recs[j].size);
        ASSERT_EQ(
            memcmp(fetched_record->data, recs[j].data, fetched_record->size),
            0);
    }

    // 事务2：应不存在
    for (size_t j = recs.size() / 2; j < recs.size(); j++) {
        try {
            auto fetched_record = fh->get_record(rids[j], new_context.get());
            FAIL() << "Record should not exist after undo.";
        } catch (RecordNotFoundError&) {
            // 预期行为，记录不应存在
        }
    }
}