/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "log_recovery.h"
#include "spdlog/spdlog.h"

std::atomic<bool> enable_logging = false;

/**
 * @description: analyze阶段，需要获得脏页表（DPT）和未完成的事务列表（ATT）
 */
void RecoveryManager::analyze() {
    // spdlog::info("=================recovery: start analyze=============");
    char log_header_buf[LOG_HEADER_SIZE + 1];
    lsn_t cur_lsn = 0;
    int log_offset = 0;

    while (disk_manager_->read_log(log_header_buf, LOG_HEADER_SIZE,
                                   log_offset) > 0) {
        std::shared_ptr<LogRecord> log_header = std::make_shared<LogRecord>();
        log_header->deserialize(log_header_buf);
        cur_lsn = log_header->lsn_;

        switch (log_header->log_type_) {
            case LogType::begin: {
                active_txn_[log_header->log_tid_] = log_header->lsn_;
                lsn_log_map_[cur_lsn] = log_header;
                break;
            }
            case LogType::commit: {
                active_txn_.erase(log_header->log_tid_);
                lsn_log_map_[cur_lsn] = log_header;
                break;
            }
            case LogType::ABORT: {
                active_txn_.erase(log_header->log_tid_);
                lsn_log_map_[cur_lsn] = log_header;
                break;
            }
            default: {
                char log_data[log_header->log_tot_len_ + 1];
                disk_manager_->read_log(log_data, log_header->log_tot_len_,
                                        log_offset);
                // TODO: 把重复逻辑尝试抽出来，可能用智能指针去类型转换
                auto log_record = std::make_shared<LogRecord>();

                if (log_header->log_type_ == LogType::INSERT) {
                    log_record = std::make_shared<InsertLogRecord>();
                } else if (log_header->log_type_ == LogType::DELETE) {
                    log_record = std::make_shared<DeleteLogRecord>();
                } else if (log_header->log_type_ == LogType::UPDATE) {
                    log_record = std::make_shared<UpdateLogRecord>();
                }
                log_record->deserialize(log_data);
                active_txn_[log_record->log_tid_] = log_record->lsn_;
                if (dirty_page_tab_.count(log_record->get_rid().page_no) == 0) {
                    dirty_page_tab_[log_record->get_rid().page_no] =
                        log_record->lsn_;
                }
                lsn_log_map_[cur_lsn] = log_record;
                break;
            }
        }
        // log_header->format_print();
        persist_lsn_ = std::max(persist_lsn_, cur_lsn);
        log_offset += log_header->log_tot_len_;
    }
}

/**
 * @description: 重做所有未落盘的操作
 */
void RecoveryManager::redo() {
    // spdlog::info("=================recovery: start redo============");
    enable_logging = false;

    if (dirty_page_tab_.empty()) {
        return;
    }
    // 从dpt中找到最早未刷新到磁盘的LSN，然后从这个LSN开始redo
    int min_lsn = dirty_page_tab_.begin()->second;
    for (auto& [page_id, lsn] : dirty_page_tab_) {
        min_lsn = std::min(min_lsn, lsn);
    }
    int lsn = min_lsn;

    while (lsn_log_map_.count(lsn) > 0) {
        auto log_record = lsn_log_map_[lsn];
        // log_record->format_print();
        switch (log_record->log_type_) {
            case LogType::begin:
            case LogType::commit:
            case LogType::ABORT:
                // maybe do nothing here in those three cases
                break;
            case LogType::INSERT: {
                auto insert_log_record =
                    std::dynamic_pointer_cast<InsertLogRecord>(log_record);
                // ! 注意context要传nullptr，避免在redo这里加锁；同时注意避免redo这里的insert_record 再写一次日志
                redo_insert(insert_log_record->get_table_name(),
                            insert_log_record->rid_,
                            insert_log_record->insert_value_, nullptr);
                break;
            }
            case LogType::DELETE: {
                auto delete_log_record =
                    std::dynamic_pointer_cast<DeleteLogRecord>(log_record);

                redo_delete(delete_log_record->get_table_name(),
                            delete_log_record->rid_,
                            delete_log_record->delete_value_, nullptr);
                break;
            }
            case LogType::UPDATE: {
                auto update_log_record =
                    std::dynamic_pointer_cast<UpdateLogRecord>(log_record);
                redo_update(update_log_record->get_table_name(),
                            update_log_record->rid_,
                            update_log_record->new_value_,
                            update_log_record->old_value_, nullptr);
                break;
            }
        }
        lsn++;
    }
}

lsn_t RecoveryManager::get_max_active_lsn() {
    lsn_t max_lsn = INVALID_LSN;
    for (auto& [tid, lsn] : active_txn_) {
        if (lsn > max_lsn) {
            max_lsn = lsn;
        }
    }
    return max_lsn;
}

/**
 * @description: 回滚未完成的事务
 */
void RecoveryManager::undo() {
    while (!active_txn_.empty()) {
        auto lsn = get_max_active_lsn();
        auto log_record = lsn_log_map_[lsn];
        bool while_flag = true;

        while (while_flag) {
            switch (log_record->log_type_) {
                case LogType::begin: {
                    active_txn_.erase(log_record->log_tid_);
                    while_flag = false;
                    break;
                }
                case LogType::commit: {
                    assert(false);
                }
                case LogType::ABORT: {
                    break;
                }
                case LogType::INSERT: {
                    auto insert_log_record =
                        std::dynamic_pointer_cast<InsertLogRecord>(log_record);
                    redo_delete(insert_log_record->get_table_name(),
                                insert_log_record->rid_,
                                insert_log_record->insert_value_, nullptr);
                    break;
                }
                case LogType::DELETE: {
                    auto delete_log_record =
                        std::dynamic_pointer_cast<DeleteLogRecord>(log_record);

                    redo_insert(delete_log_record->get_table_name(),
                                delete_log_record->rid_,
                                delete_log_record->delete_value_, nullptr);
                    break;
                }
                case LogType::UPDATE: {
                    auto update_log_record =
                        std::dynamic_pointer_cast<UpdateLogRecord>(log_record);
                    redo_update(update_log_record->get_table_name(),
                                update_log_record->rid_,
                                update_log_record->old_value_,
                                update_log_record->new_value_, nullptr);
                    break;
                }
            }
            if (log_record->prev_lsn_ == INVALID_LSN) {
                active_txn_.erase(log_record->log_tid_);
                break;
            } else {
                active_txn_[log_record->log_tid_] = log_record->prev_lsn_;
                // spdlog::info("undo: tid: {}, lsn: {}, prev_lsn: {}",
                //              log_record->log_tid_, lsn, log_record->prev_lsn_);
                lsn = log_record->prev_lsn_;
                log_record = lsn_log_map_[lsn];
            }
        }
    }

    // ! check: undo 结束是否自动调用 finish_job()
    finish_job();
}

void RecoveryManager::finish_job() {
    // redo,undo 搞定之后马上做一次数据落盘，保证鲁棒性
    sm_manager_->flush_data();

    // 清空数据结构
    active_txn_.clear();
    dirty_page_tab_.clear();
    lsn_log_map_.clear();

    // 开启日志记录
    enable_logging = true;
}

void RecoveryManager::redo_insert(const std::string& tab_name, const Rid& rid,
                                  const RmRecord& record, Transaction* txn) {
    auto fh = sm_manager_->fhs_[tab_name].get();
    auto& tab_meta = sm_manager_->db_.get_table(tab_name);
    // 先插入记录再插入索引
    fh->insert_record(rid, record.data, nullptr);
    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(record.data);
        ih->insert_entry(key.get(), rid, txn);
    }
}

void RecoveryManager::redo_delete(const std::string& tab_name, const Rid& rid,
                                  const RmRecord& record, Transaction* txn) {
    auto fh = sm_manager_->fhs_[tab_name].get();
    auto& tab_meta = sm_manager_->db_.get_table(tab_name);
    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(record.data);
        ih->delete_entry(key.get(), txn);
    }
    fh->delete_record(rid, nullptr);
}

void RecoveryManager::redo_update(const std::string& tab_name, const Rid& rid,
                                  const RmRecord& record,
                                  const RmRecord& old_record,
                                  Transaction* txn) {
    // 这里等价于 rollback_update(tab_name, rid, old_record, record, txn);

    auto fh = sm_manager_->fhs_[tab_name].get();
    auto& tab_meta = sm_manager_->db_.get_table(tab_name);
    // 删除旧索引，用新的记录更新，插入新索引
    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(old_record.data);
        ih->delete_entry(key.get(), txn);
    }

    fh->update_record(rid, record.data, nullptr);

    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(record.data);
        ih->insert_entry(key.get(), rid, txn);
    }
}
