/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "sm_manager.h"

#include <sys/stat.h>
#include <unistd.h>

#include <fstream>

#include "common/config.h"
#include "errors.h"
#include "index/ix.h"
#include "record/rm.h"
#include "record/rm_file_handle.h"
#include "record/rm_scan.h"
#include "record_printer.h"

/**
 * @description: 判断是否为一个文件夹
 * @return {bool} 返回是否为一个文件夹
 * @param {string&} db_name 数据库文件名称，与文件夹同名
 */
bool SmManager::is_dir(const std::string& db_name) {
    struct stat st;
    return stat(db_name.c_str(), &st) == 0 && S_ISDIR(st.st_mode);
}

/**
 * @description: 创建数据库，所有的数据库相关文件都放在数据库同名文件夹下
 * @param {string&} db_name 数据库名称
 */
void SmManager::create_db(const std::string& db_name) {
    if (is_dir(db_name)) {
        throw DatabaseExistsError(db_name);
    }
    //为数据库创建一个子目录
    std::string cmd = "mkdir " + db_name;
    if (system(cmd.c_str()) < 0) {  // 创建一个名为db_name的目录
        throw UnixError();
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    //创建系统目录
    DbMeta* new_db = new DbMeta();
    new_db->name_ = db_name;

    // 注意，此处ofstream会在当前目录创建(如果没有此文件先创建)和打开一个名为DB_META_NAME的文件
    std::ofstream ofs(DB_META_NAME);

    // 将new_db中的信息，按照定义好的operator<<操作符，写入到ofs打开的DB_META_NAME文件中
    ofs << *new_db;  // 注意：此处重载了操作符<<

    delete new_db;

    // 创建日志文件
    disk_manager_->create_file(LOG_FILE_NAME);

    // 回到根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 删除数据库，同时需要清空相关文件以及数据库同名文件夹
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::drop_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    std::string cmd = "rm -r " + db_name;
    if (system(cmd.c_str()) < 0) {
        throw UnixError();
    }
}

/**
 * @description: 打开数据库，找到数据库对应的文件夹，并加载数据库元数据和相关文件
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::open_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    std::ifstream ifs(DB_META_NAME);
    if (!ifs.is_open()) {
        std::cout << "[ERROR] Failed to open " << DB_META_NAME << std::endl;
        throw UnixError();
    }
    ifs >> db_;
    ifs.close();
    for (auto& tab_entry : db_.tabs_) {
        auto& tab = tab_entry.second;
        fhs_[tab.name] = rm_manager_->open_file(tab.name);

        for (auto& index : tab.indexes) {
            auto index_handle = ix_manager_->open_index(tab.name, index.cols);
            auto index_name = ix_manager_->get_index_name(tab.name, index.cols);
            ihs_.emplace(index_name, std::move(index_handle));
        }
    }
}

/**
 * @description: 把数据库相关的元数据刷入磁盘中
 */
void SmManager::flush_meta() {
    // 默认清空文件
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
}

void SmManager::flush_data() {

    for (auto& tab_entry : db_.tabs_) {
        auto& tab = tab_entry.second;

        rm_manager_->flush_file(fhs_[tab.name].get());

        for (auto& index : tab.indexes) {
            auto index_name = ix_manager_->get_index_name(tab.name, index.cols);
            ix_manager_->flush_index(ihs_[index_name].get());
        }
    }
}

/**
 * @description: 关闭数据库并把数据落盘
 */
void SmManager::close_db() {
    // 目前在数据库文件夹中
    for (auto& tab_entry : db_.tabs_) {
        auto& tab = tab_entry.second;
        rm_manager_->close_file(fhs_[tab.name].get());

        for (auto& index : tab.indexes) {
            auto index_name = ix_manager_->get_index_name(tab.name, index.cols);
            ix_manager_->close_index(ihs_[index_name].get());
        }
    }

    flush_meta();
    fhs_.clear();
    ihs_.clear();

    db_.name_.clear();
    db_.tabs_.clear();
    // 回到上级的根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 显示所有的表,通过测试需要将其结果写入到output.txt,详情看题目文档
 * @param {Context*} context 
 */
void SmManager::show_tables(Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    outfile << "| Tables |\n";
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"Tables"}, context);
    printer.print_separator(context);
    for (auto& entry : db_.tabs_) {
        auto& tab = entry.second;
        printer.print_record({tab.name}, context);
        outfile << "| " << tab.name << " |\n";
    }
    printer.print_separator(context);
    outfile.close();
}

/**
 * @description: 显示表的元数据
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::desc_table(const std::string& tab_name, Context* context) {
    TabMeta& tab = db_.get_table(tab_name);

    std::vector<std::string> captions = {"Field", "Type", "Index"};
    RecordPrinter printer(captions.size());
    // Print header
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);
    // Print fields
    for (auto& col : tab.cols) {
        std::vector<std::string> field_info = {col.name, coltype2str(col.type),
                                               col.index ? "YES" : "NO"};
        printer.print_record(field_info, context);
    }
    // Print footer
    printer.print_separator(context);
}

/**
 * @description: 创建表
 * @param {string&} tab_name 表的名称
 * @param {vector<ColDef>&} col_defs 表的字段
 * @param {Context*} context 
 */
void SmManager::create_table(const std::string& tab_name,
                             const std::vector<ColDef>& col_defs,
                             Context* context) {
    if (db_.is_table(tab_name)) {
        throw TableExistsError(tab_name);
    }
    // Create table meta
    int curr_offset = 0;
    TabMeta tab;
    tab.name = tab_name;
    for (auto& col_def : col_defs) {
        ColMeta col = {.tab_name = tab_name,
                       .name = col_def.name,
                       .type = col_def.type,
                       .len = col_def.len,
                       .offset = curr_offset,
                       .index = false};
        curr_offset += col_def.len;
        tab.cols.push_back(col);
    }
    // Create & open record file
    int record_size =
        curr_offset;  // record_size就是col meta所占的大小（表的元数据也是以记录的形式进行存储的）
    rm_manager_->create_file(tab_name, record_size);
    db_.tabs_[tab_name] = tab;
    // fhs_[tab_name] = rm_manager_->open_file(tab_name);
    fhs_.emplace(tab_name, rm_manager_->open_file(tab_name));

    context->lock_mgr_->lock_exclusive_on_table(context->txn_,
                                                fhs_.at(tab_name)->GetFd());
    flush_meta();
}

/**
 * @description: 删除表
 * @param {string&} tab_name 表的名称
 * @param {Context*} context
 */
void SmManager::drop_table(const std::string& tab_name, Context* context) {
    /** TODO: 
        1. 事务影响
    */
    /** STEP:
     * 1. 
     */
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    context->lock_mgr_->lock_exclusive_on_table(context->txn_,
                                                fhs_.at(tab_name)->GetFd());
    // 删除索引
    auto& tab = db_.get_table(tab_name);
    for (auto& index : tab.indexes) {
        auto index_name = ix_manager_->get_index_name(tab_name, index.cols);
        drop_index(tab_name, index.cols, context);
    }

    if (fhs_.find(tab_name) != fhs_.end()) {
        rm_manager_->close_file(fhs_[tab_name].get());
        fhs_.erase(tab_name);
    }
    db_.tabs_.erase(tab_name);
    rm_manager_->destroy_file(tab_name);

    flush_meta();
}

/**
 * @description: 创建索引
 * @param {string&} tab_name 表的名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::create_index(const std::string& tab_name,
                             const std::vector<std::string>& col_names,
                             Context* context) {
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    if (ix_manager_->exists(tab_name, col_names)) {
        throw IndexExistsError(tab_name, col_names);
    }

    context->lock_mgr_->lock_shared_on_table(context->txn_,
                                             fhs_.at(tab_name)->GetFd());

    auto& tab = db_.get_table(tab_name);

    IndexMeta index_meta;
    index_meta.tab_name = tab_name, index_meta.col_tot_len = 0;
    std::vector<ColMeta> index_cols;
    for (auto& col_name : col_names) {
        auto col = tab.get_col(col_name);
        index_cols.push_back(*col);  // 获取索引列字段的meta
        index_meta.col_tot_len += col->len;
    }
    index_meta.cols = index_cols, index_meta.col_num = index_cols.size();

    // 创建并打开索引文件
    ix_manager_->create_index(tab_name, index_cols);
    auto index_handle = ix_manager_->open_index(tab_name, index_cols);

    auto index_name = ix_manager_->get_index_name(tab_name, col_names);
    tab.indexes.push_back(index_meta);

    // 为现有的记录 insert index entry
    auto file_handle = fhs_[tab_name].get();
    RmScan rm_scan(file_handle);
    while (!rm_scan.is_end()) {
        auto record = file_handle->get_record(rm_scan.rid(), context);

        auto key = index_meta.get_index_key(record->data);
        index_handle->insert_entry(key.get(), rm_scan.rid(), context->txn_);

        rm_scan.next();
    }
    // ix_manager_->close_index(index_handle.get());
    ix_manager_->flush_index(index_handle.get());
    ihs_.emplace(index_name, std::move(index_handle));
    flush_meta();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name,
                           const std::vector<std::string>& col_names,
                           Context* context) {
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    context->lock_mgr_->lock_exclusive_on_table(context->txn_,
                                                fhs_.at(tab_name)->GetFd());

    auto& tab = db_.get_table(tab_name);

    if (!ix_manager_->exists(tab_name, col_names)) {
        throw IndexNotFoundError(tab_name, col_names);
    }

    auto index_name = ix_manager_->get_index_name(tab_name, col_names);
    ix_manager_->close_index(ihs_[index_name].get());
    ix_manager_->destroy_index(tab_name, col_names);

    auto index_meta = tab.get_index_meta(col_names);
    tab.indexes.erase(index_meta);
    ihs_.erase(index_name);
    flush_meta();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<ColMeta>&} 索引包含的字段元数据
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name,
                           const std::vector<ColMeta>& cols, Context* context) {
    std::vector<std::string> col_names;
    for (auto& col : cols) {
        col_names.push_back(col.name);
    }
    drop_index(tab_name, col_names, context);
}

void SmManager::show_index(const std::string& tab_name, Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);

    RecordPrinter printer(3);
    printer.print_separator(context);
    for (auto& index : db_.get_table(tab_name).indexes) {
        std::vector<std::string> index_cols;
        index_cols.push_back(tab_name);
        index_cols.push_back("unique");

        std::string index_col = "(";
        outfile << "| " << tab_name << " | unique | (";
        for (auto it = index.cols.begin(); it != index.cols.end() - 1; it++) {
            outfile << it->name << ",";
            index_col = index_col + it->name + ",";
        }
        outfile << index.cols.back().name << ") |\n";
        index_col = index_col + index.cols.back().name + ")";
        index_cols.push_back(index_col);
        printer.print_record(index_cols, context);
    }
    printer.print_separator(context);
    outfile.close();
}
