/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"
#include <cassert>
#include "transaction/txn_defs.h"

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid,
                                        int tab_fd) {
    LockDataId lock_data_id(tab_fd, rid, LockDataType::RECORD);
    return lock(txn, LockMode::SHARED, lock_data_id);
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid,
                                           int tab_fd) {

    LockDataId lock_data_id(tab_fd, rid, LockDataType::RECORD);
    return lock(txn, LockMode::EXCLUSIVE, lock_data_id);
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    return lock(txn, LockMode::SHARED, lock_data_id);
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    return lock(txn, LockMode::EXCLUSIVE, lock_data_id);
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    return lock(txn, LockMode::INTENTION_SHARED, lock_data_id);
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    return lock(txn, LockMode::INTENTION_EXCLUSIVE, lock_data_id);
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {
    if (txn->get_state() == TransactionState::GROWING) {
        txn->set_state(TransactionState::SHRINKING);
    }

    std::unique_lock<std::mutex> manager_lock(latch_);
    // 如果锁表中还不存在该对象，会自动默认构造一个新的请求队列，因此这里不用再构造了，直接哈希表[]访问即可
    auto& lock_request_queue = lock_table_[lock_data_id];
    std::unique_lock<std::mutex> rq_lock(lock_request_queue.latch_);
    manager_lock.unlock();

    auto it =
        std::find_if(lock_request_queue.request_queue_.begin(),
                     lock_request_queue.request_queue_.end(),
                     [txn](const LockRequest& request) {
                         return request.txn_id_ == txn->get_transaction_id();
                     });
    LockMode unlock_mode;
    if (it != lock_request_queue.request_queue_.end()) {
        assert(it->granted_);
        remove_lock_set(txn, lock_data_id, it->lock_mode_);
        unlock_mode = it->lock_mode_;
        lock_request_queue.request_queue_.erase(it);
        // !check: 更新group_lock_mode_，是否可以直接更新为non_lock，还是需要再遍历一遍request_queue_
        lock_request_queue.group_lock_mode_ =
            update_group_lock(lock_request_queue, lock_data_id);
        // lock_request_queue.group_lock_mode_ = GroupLockMode::NON_LOCK;
    } else {
        return false;
    }

    // 通知所有等待该锁的事务
    lock_request_queue.cv_.notify_all();
    spdlog::info("txn {} unlock {}_lock {}", (int)txn->get_transaction_id(),
                 lock_mode_to_str(unlock_mode), lock_data_id.ToString());
    return true;
}

bool LockManager::lock(Transaction* txn, LockMode lock_mode,
                       LockDataId lock_data_id) {
    // 1. 检查事务状态
    int txn_id = txn->get_transaction_id();
    if (!check_txn_state(txn)) {
        return false;
    }

    // 2. 检查锁状态，顺便检查是否可以升级
    bool is_upgrade = false, txn_has_lock = false;
    // 2.1. 如果已经有事务申请了相同或更强的锁，则直接返回成功
    // 2.2. 如果申请了更弱的锁，则需要考虑锁升级
    if (check_lock_state(txn, lock_data_id, lock_mode, lock_data_id.type_,
                         is_upgrade)) {
        // TODO: 修改这个log，这里实际没加锁
        // print_lock_log(txn, lock_data_id, lock_mode);
        return true;
    }

    // 3. 获取锁请求队列
    // 先对锁管理器加锁，加锁后从锁表中获取请求队列，再对请求队列加锁，锁管理器解锁
    std::unique_lock<std::mutex> manager_lock(latch_);
    // 如果锁表中还不存在该对象，会自动默认构造一个新的请求队列，因此这里不用再构造了，直接哈希表[]访问即可
    auto& lock_request_queue = lock_table_[lock_data_id];
    std::unique_lock<std::mutex> rq_lock(lock_request_queue.latch_);
    manager_lock.unlock();

    // 4. 如果能升级，尝试升级锁
    if (is_upgrade == true) {
        // 4.1. 如果有其他事务正在锁升级，升级冲突
        if (lock_request_queue.upgrading_) {
            txn->set_state(TransactionState::ABORTED);
            throw TransactionAbortException(txn_id,
                                            AbortReason::UPGRADE_CONFLICT);
        }
        for (auto req_it = lock_request_queue.request_queue_.begin();
             req_it != lock_request_queue.request_queue_.end(); ++req_it) {
            if (req_it->txn_id_ != txn_id || !req_it->granted_) {
                continue;
            }
            txn_has_lock = true;
            lock_request_queue.upgrading_ = true;
            // 4.2 锁升级操作：检查升级兼容性，返回升级后的锁模式upd_mode
            LockMode old_mode, upd_mode;
            if (check_update_compatibility(lock_data_id, req_it->lock_mode_,
                                           lock_mode, upd_mode)) {
                // 4.3. 尝试升级锁，check_update_compatibility() 会给出升级后的锁模式
                while (!check_queue_lock_compatibility(lock_request_queue,
                                                       upd_mode, txn)) {
                    // 4.3.1. 如果当前该数据项上持有的锁与即将升级的锁不兼容
                    // 4.3.2 如果现在的事务较晚开始，优先级比不上之前的事务，则直接返回失败 (txn_id < old_lock_request.txn_id_)
                    if (!check_wait_die(txn, &lock_request_queue, upd_mode)) {
                        // ! check the condition here
                        txn->set_state(TransactionState::ABORTED);
                        lock_request_queue.upgrading_ = false;
                        throw TransactionAbortException(
                            txn_id, AbortReason::DEADLOCK_PREVENTION);
                    }
                    // 4.4 如果优先级是OK的，那么就等待持有锁的事务去释放锁
                    lock_request_queue.cv_.wait(rq_lock);
                }
                old_mode = req_it->lock_mode_;
                req_it->lock_mode_ = upd_mode;
                lock_request_queue.group_lock_mode_ =
                    convert_lock2group(upd_mode);
                insert_lock_set(txn, lock_data_id, upd_mode);
                remove_lock_set(txn, lock_data_id, old_mode);
                // lock_request_queue.request_queue_.erase(req_it);
                lock_request_queue.upgrading_ = false;
                print_lock_log(txn, lock_data_id, lock_mode);
                return true;
            } else {
                txn->set_state(TransactionState::ABORTED);
                lock_request_queue.upgrading_ = false;
                throw TransactionAbortException(txn_id,
                                                AbortReason::UPGRADE_CONFLICT);
            }
        }
    }

    // 5. 如果请求队列中没有txn申请的锁请求，则创建一个新的锁请求
    if (!txn_has_lock) {
        while (!check_queue_lock_compatibility(lock_request_queue, lock_mode,
                                               txn)) {
            // 5.1. 如果当前该数据项上持有的锁与申请的锁不兼容
            // 5.1.1 如果现在的事务较晚开始，优先级比不上之前的事务，则直接返回失败 (txn_id < old_lock_request.txn_id_)
            if (!check_wait_die(txn, &lock_request_queue, lock_mode)) {
                // ! check the condition here
                txn->set_state(TransactionState::ABORTED);
                throw TransactionAbortException(
                    txn_id, AbortReason::DEADLOCK_PREVENTION);
            }
            // 5.2 如果优先级是OK的，那么就等待持有锁的事务去释放锁
            lock_request_queue.cv_.wait(rq_lock);
        }
        LockRequest new_request(txn_id, lock_mode);
        new_request.granted_ = true;
        lock_request_queue.request_queue_.push_back(new_request);
        lock_request_queue.group_lock_mode_ = convert_lock2group(lock_mode);
        insert_lock_set(txn, lock_data_id, lock_mode);
    }

    print_lock_log(txn, lock_data_id, lock_mode);
    return true;
}

void LockManager::unlock_all_locks(Transaction* txn) {
    while (!txn->get_s_row_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_s_row_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
    while (!txn->get_x_row_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_x_row_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
    while (!txn->get_s_tab_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_s_tab_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
    while (!txn->get_x_tab_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_x_tab_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
    while (!txn->get_is_tab_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_is_tab_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
    while (!txn->get_ix_tab_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_ix_tab_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
    while (!txn->get_six_tab_lock_set()->empty()) {
        LockDataId lock_data_id = *txn->get_six_tab_lock_set()->begin();
        unlock(txn, lock_data_id);
    }
}

void LockManager::insert_lock_set(Transaction* txn, LockDataId lock_data_id,
                                  LockMode lock_mode) {
    if (lock_data_id.type_ == LockDataType::TABLE) {
        switch (lock_mode) {
            case LockMode::SHARED:
                txn->get_s_tab_lock_set()->insert(lock_data_id);
                break;
            case LockMode::EXCLUSIVE:
                txn->get_x_tab_lock_set()->insert(lock_data_id);
                break;

            case LockMode::INTENTION_SHARED:
                txn->get_is_tab_lock_set()->insert(lock_data_id);
                break;
            case LockMode::INTENTION_EXCLUSIVE:
                txn->get_ix_tab_lock_set()->insert(lock_data_id);
                break;
            case LockMode::S_IX:
                txn->get_six_tab_lock_set()->insert(lock_data_id);
                break;
            default:
                break;
        }
    } else if (lock_data_id.type_ == LockDataType::RECORD) {
        switch (lock_mode) {
            case LockMode::SHARED:
                txn->get_s_row_lock_set()->insert(lock_data_id);
                break;
            case LockMode::EXCLUSIVE:
                txn->get_x_row_lock_set()->insert(lock_data_id);
                break;
            default:
                // TODO: maybe throw an exception here
                assert(false);
                break;
        }
    }
}

void LockManager::remove_lock_set(Transaction* txn, LockDataId lock_data_id,
                                  LockMode lock_mode) {
    if (lock_data_id.type_ == LockDataType::TABLE) {
        switch (lock_mode) {
            case LockMode::SHARED:
                txn->get_s_tab_lock_set()->erase(lock_data_id);
                break;
            case LockMode::EXCLUSIVE:
                txn->get_x_tab_lock_set()->erase(lock_data_id);
                break;

            case LockMode::INTENTION_SHARED:
                txn->get_is_tab_lock_set()->erase(lock_data_id);
                break;
            case LockMode::INTENTION_EXCLUSIVE:
                txn->get_ix_tab_lock_set()->erase(lock_data_id);
                break;
            case LockMode::S_IX:
                txn->get_six_tab_lock_set()->erase(lock_data_id);
                break;
            default:
                break;
        }
    } else if (lock_data_id.type_ == LockDataType::RECORD) {
        switch (lock_mode) {
            case LockMode::SHARED:
                txn->get_s_row_lock_set()->erase(lock_data_id);
                break;
            case LockMode::EXCLUSIVE:
                txn->get_x_row_lock_set()->erase(lock_data_id);
                break;
            default:
                // TODO: maybe throw an exception here
                assert(false);
                break;
        }
    }
}

bool LockManager::check_update_compatibility(LockDataId lock, LockMode old_mode,
                                             LockMode new_mode,
                                             LockMode& res_mode) {
    // 需要区分升级的是表锁还是行锁
    if (lock.type_ == LockDataType::TABLE) {
        switch (old_mode) {
            case LockMode::INTENTION_SHARED:
                if (new_mode == LockMode::INTENTION_SHARED) {
                    return false;
                }
                // IS + IX = IX, IS + S = S, IS + X = X
                res_mode = new_mode;
                break;
            case LockMode::INTENTION_EXCLUSIVE:
                if (new_mode == LockMode::EXCLUSIVE) {
                    // IX + X = X
                    res_mode = LockMode::EXCLUSIVE;
                } else if (new_mode == LockMode::SHARED) {
                    // IX + S = S_IX
                    res_mode = LockMode::S_IX;
                } else {
                    return false;
                }
                break;
            case LockMode::S_IX:
                if (new_mode != LockMode::EXCLUSIVE) {
                    // SIX + X = SIX
                    return false;
                }
                res_mode = LockMode::EXCLUSIVE;
                break;
            case LockMode::SHARED:
                if (new_mode == LockMode::INTENTION_EXCLUSIVE) {
                    // S + IX = S_IX
                    res_mode = LockMode::S_IX;
                } else if (new_mode == LockMode::EXCLUSIVE) {
                    // S + X = X
                    res_mode = LockMode::EXCLUSIVE;
                } else {
                    return false;
                }
                break;
            case LockMode::EXCLUSIVE:
                return false;
            default:
                return false;
        }
        return true;
    } else if (lock.type_ == LockDataType::RECORD) {
        switch (old_mode) {
            case LockMode::SHARED:
                if (new_mode == LockMode::SHARED) {
                    return false;
                }
                // S + X = X, S + S = S
                res_mode = new_mode;
                break;
            case LockMode::EXCLUSIVE:
                if (new_mode == LockMode::EXCLUSIVE) {
                    return false;
                }
                // X + X = X
                res_mode = LockMode::EXCLUSIVE;
                break;
            default:
                return false;
        }
        return true;
    }
    return true;
}

bool LockManager::check_lock_compatibility(GroupLockMode group_mode,
                                           LockMode new_mode) const {
    switch (group_mode) {
        case GroupLockMode::NON_LOCK:
            return true;
        case GroupLockMode::IS:
            if (new_mode == LockMode::EXCLUSIVE) {
                return false;
            }
            break;
        case GroupLockMode::IX:
            if (new_mode != LockMode::INTENTION_EXCLUSIVE &&
                new_mode != LockMode::INTENTION_SHARED) {
                return false;
            }
            break;
        case GroupLockMode::SIX:
            if (new_mode != LockMode::INTENTION_SHARED) {
                return false;
            }
        case GroupLockMode::S:
            if (new_mode != LockMode::SHARED &&
                new_mode != LockMode::INTENTION_SHARED) {
                return false;
            }
            break;
        case GroupLockMode::X:
            return false;
        default:
            return false;
    }
    return true;
}

LockManager::GroupLockMode LockManager::convert_lock2group(LockMode lock_mode) {
    switch (lock_mode) {
        case LockMode::INTENTION_SHARED:
            return GroupLockMode::IS;
        case LockMode::INTENTION_EXCLUSIVE:
            return GroupLockMode::IX;
        case LockMode::SHARED:
            return GroupLockMode::S;
        case LockMode::S_IX:
            return GroupLockMode::SIX;
        case LockMode::EXCLUSIVE:
            return GroupLockMode::X;
        default:
            return GroupLockMode::NON_LOCK;
    }
}

LockManager::LockMode LockManager::convert_group2lock(
    GroupLockMode group_lock_mode) {
    switch (group_lock_mode) {
        case GroupLockMode::IS:
            return LockMode::INTENTION_SHARED;
        case GroupLockMode::IX:
            return LockMode::INTENTION_EXCLUSIVE;
        case GroupLockMode::S:
            return LockMode::SHARED;
        case GroupLockMode::SIX:
            return LockMode::S_IX;
        case GroupLockMode::X:
            return LockMode::EXCLUSIVE;
        default:
            assert(false);
    }
}

bool LockManager::check_queue_lock_compatibility(
    LockRequestQueue& lock_request_queue, LockMode new_mode,
    Transaction* txn) const {
    auto txn_id = txn->get_transaction_id();
    for (auto req : lock_request_queue.request_queue_) {
        // 只检查别的事务被授予的锁；目前遍历一遍的方法比较笨，考虑后续优化
        if (req.txn_id_ == txn_id || !req.granted_) {
            continue;
        }
        if (!check_lock_compatibility(convert_lock2group(req.lock_mode_),
                                      new_mode)) {
            return false;
        }
    }
    return true;
}

bool LockManager::check_lock_state(Transaction* txn, LockDataId lock_data_id,
                                   LockMode new_mode, LockDataType lock_type,
                                   bool& is_upgrade) const {
    // TODO: Maybe add check to DEADLOCK_PREVENTION
    if (lock_type == LockDataType::TABLE) {
        switch (new_mode) {
            case LockMode::INTENTION_SHARED: {
                if (txn->is_tab_is_locked(lock_data_id) ||
                    txn->is_tab_ix_locked(lock_data_id) ||
                    txn->is_tab_six_locked(lock_data_id) ||
                    txn->is_tab_x_locked(lock_data_id) ||
                    txn->is_tab_s_locked(lock_data_id)) {
                    return true;
                }
                // 什么锁都没有，无法升级
                is_upgrade = false;
                break;
            }
            case LockMode::INTENTION_EXCLUSIVE: {
                if (txn->is_tab_ix_locked(lock_data_id) ||
                    txn->is_tab_six_locked(lock_data_id) ||
                    txn->is_tab_x_locked(lock_data_id)) {
                    // !check
                    return true;
                } else if (txn->is_tab_is_locked(lock_data_id) ||
                           txn->is_tab_s_locked(lock_data_id)) {
                    // IX + S = SIX, IX + IS = IX
                    is_upgrade = true;
                }
                // 什么锁都没有，无法升级
                is_upgrade = false;
                break;
            }
            case LockMode::SHARED: {
                if (txn->is_tab_s_locked(lock_data_id) ||
                    txn->is_tab_x_locked(lock_data_id) ||
                    txn->is_tab_six_locked(lock_data_id)) {
                    // !check
                    return true;
                } else if (txn->is_tab_is_locked(lock_data_id) ||
                           txn->is_tab_ix_locked(lock_data_id)) {
                    // S + IX = SIX, S + IS = S
                    is_upgrade = true;
                }
                // 什么锁都没有，无法升级
                is_upgrade = false;
                break;
            }
            case LockMode::S_IX: {
                // SIX 一定是升级来的，不会显式的加锁
                assert(false);
                if (txn->is_tab_six_locked(lock_data_id) ||
                    txn->is_tab_x_locked(lock_data_id)) {
                    return true;
                }
                is_upgrade = false;
                break;
            }
            case LockMode::EXCLUSIVE: {
                if (txn->is_tab_x_locked(lock_data_id)) {
                    return true;
                } else if (txn->is_tab_s_locked(lock_data_id) ||
                           txn->is_tab_is_locked(lock_data_id) ||
                           txn->is_tab_ix_locked(lock_data_id) ||
                           txn->is_tab_six_locked(lock_data_id)) {
                    is_upgrade = true;
                }
                // 什么锁都没有，无法升级
                is_upgrade = false;
                break;
            }
        }
    } else if (lock_type == LockDataType::RECORD) {
        if (new_mode == LockMode::SHARED) {
            if (txn->is_row_s_locked(lock_data_id) ||
                txn->is_row_x_locked(lock_data_id)) {
                return true;
            }
            is_upgrade = false;
        } else if (new_mode == LockMode::EXCLUSIVE) {
            if (txn->is_row_x_locked(lock_data_id)) {
                return true;
            } else if (txn->is_row_s_locked(lock_data_id)) {
                is_upgrade = true;
                return false;
            }
        }
    }

    return false;
}

bool LockManager::check_wait_die(Transaction* txn,
                                 LockRequestQueue* lock_request_queue,
                                 LockMode lock_mode) {
    auto txn_id = txn->get_transaction_id();
    for (auto req : lock_request_queue->request_queue_) {
        if (req.granted_ && req.txn_id_ != txn_id) {
            auto old_group_mode = convert_lock2group(req.lock_mode_);
            if (check_lock_compatibility(old_group_mode, lock_mode)) {
                // 只有不兼容的锁才会阻塞
                continue;
            }
            if (req.txn_id_ < txn_id) {
                return false;
            }
        }
    }
    return true;
}

bool LockManager::check_txn_state(Transaction* txn) const {
    if (txn->get_state() == TransactionState::COMMITTED ||
        txn->get_state() == TransactionState::ABORTED) {
        return false;
    } else if (txn->get_state() == TransactionState::GROWING) {
        return true;
    } else if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
        return true;
    }
    // state == shrinking
    throw TransactionAbortException(txn->get_transaction_id(),
                                    AbortReason::LOCK_ON_SHIRINKING);
    return false;
}

LockManager::GroupLockMode LockManager::update_group_lock(
    LockRequestQueue& lock_request_queue, LockDataId lock) {
    // TODO: 这里的实现有点奇怪，先不改了；比较优化的实现是“锁降级”的过程
    GroupLockMode mode = GroupLockMode::NON_LOCK;
    LockMode upd_mode;
    for (auto& req : lock_request_queue.request_queue_) {
        if (!req.granted_) {
            continue;
        }
        if (mode == GroupLockMode::NON_LOCK) {
            mode = convert_lock2group(req.lock_mode_);
            upd_mode = convert_group2lock(mode);
        } else {
            // mode != NON_LOCK
            check_update_compatibility(lock, convert_group2lock(mode),
                                       req.lock_mode_, upd_mode);
        }
    }
    return (mode == GroupLockMode::NON_LOCK) ? mode
                                             : convert_lock2group(upd_mode);
}

std::string LockManager::lock_mode_to_str(LockMode lock_mode) {
    switch (lock_mode) {
        case LockMode::SHARED:
            return "S";
        case LockMode::EXCLUSIVE:
            return "X";
        case LockMode::INTENTION_SHARED:
            return "IS";
        case LockMode::INTENTION_EXCLUSIVE:
            return "IX";
        case LockMode::S_IX:
            return "SIX";
        default:
            return "UNKNOWN";
    }
}

void LockManager::print_lock_log(Transaction* txn, LockDataId lock_data_id,
                                 LockMode lock_mode) {
    std::string lock_mode_str = lock_mode_to_str(lock_mode);
    spdlog::info("txn {} lock {}_lock {}", txn->get_transaction_id(),
                 lock_mode_str, lock_data_id.ToString());
}
