/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <algorithm>
#include <cassert>
#include <condition_variable>
#include <mutex>
#include "spdlog/spdlog.h"
#include "transaction/transaction.h"
#include "transaction/txn_defs.h"

static const std::string GroupLockModeStr[10] = {"NON_LOCK", "IS", "IX",
                                                 "S",        "X",  "SIX"};

class LockManager {
    /* 加锁类型，包括共享锁、排他锁、意向共享锁、意向排他锁、SIX（意向排他锁+共享锁） */
    enum class LockMode {
        SHARED,
        EXCLUSIVE,
        INTENTION_SHARED,
        INTENTION_EXCLUSIVE,
        S_IX
    };

    /* 用于标识加锁队列中排他性最强的锁类型，例如加锁队列中有SHARED和EXLUSIVE两个加锁操作，则该队列的锁模式为X */
    enum class GroupLockMode { NON_LOCK, IS, IX, S, X, SIX };

    /* 事务的加锁申请 */
    class LockRequest {
       public:
        LockRequest(txn_id_t txn_id, LockMode lock_mode)
            : txn_id_(txn_id), lock_mode_(lock_mode), granted_(false) {}

        txn_id_t txn_id_;     // 申请加锁的事务ID
        LockMode lock_mode_;  // 事务申请加锁的类型
        bool granted_;        // 该事务是否已经被赋予锁
    };

    /* 数据项上的加锁队列 */
    class LockRequestQueue {
       public:
        std::list<LockRequest> request_queue_;  // 加锁队列

        // 条件变量，用于唤醒正在等待加锁的申请，在no-wait策略下无需使用
        std::condition_variable cv_;

        // 表示现在加锁队列的锁模式组合结果，即加锁队列中排他性最强的锁类型
        GroupLockMode group_lock_mode_ = GroupLockMode::NON_LOCK;
        std::mutex latch_;        // 用于LockRequestQueue的互斥访问
        bool upgrading_ = false;  // 是否处于升级状态，用于锁的升级
    };

   public:
    LockManager() {}

    ~LockManager() {}

    bool lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd);

    bool lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd);

    bool lock_shared_on_table(Transaction* txn, int tab_fd);

    bool lock_exclusive_on_table(Transaction* txn, int tab_fd);

    bool lock_IS_on_table(Transaction* txn, int tab_fd);

    bool lock_IX_on_table(Transaction* txn, int tab_fd);

    bool lock(Transaction* txn, LockMode lock_mode, LockDataId lock_data_id);

    bool unlock(Transaction* txn, LockDataId lock_data_id);

    void unlock_all_locks(Transaction* txn);

   private:
    std::mutex latch_;  // 用于锁表的并发
    std::unordered_map<LockDataId, LockRequestQueue> lock_table_;  // 全局锁表

    void insert_lock_set(Transaction* txn, LockDataId lock_data_id,
                         LockMode lock_mode);

    void remove_lock_set(Transaction* txn, LockDataId lock_data_id,
                         LockMode lock_mode);

    bool check_update_compatibility(LockDataId lock, LockMode old_mode,
                                    LockMode new_mode, LockMode& res_mode);

    bool check_lock_compatibility(GroupLockMode group_mode,
                                  LockMode new_mode) const;

    static GroupLockMode convert_lock2group(LockMode lock_mode);

    static LockMode convert_group2lock(GroupLockMode group_lock_mode);

    bool check_queue_lock_compatibility(LockRequestQueue& lock_request_queue,
                                        LockMode new_mode,
                                        Transaction* txn) const;

    bool check_lock_state(Transaction* txn, LockDataId lock_data_id,
                          LockMode new_mode, LockDataType lock_type,
                          bool& is_upgrade) const;

    bool check_wait_die(Transaction* txn, LockRequestQueue* lock_request_queue,
                        LockMode lock_mode);

    bool check_txn_state(Transaction* txn) const;

    GroupLockMode update_group_lock(LockRequestQueue& lock_request_queue,
                                    LockDataId lock);

    std::string lock_mode_to_str(LockMode lock_mode);

    void print_lock_log(Transaction* txn, LockDataId lock_data_id,
                        LockMode lock_mode);
};
