#include "transaction/concurrency/lock_manager.h"
#include "gtest/gtest.h"
#include "index/ix_manager.h"
#include "record/rm_manager.h"
#include "system/sm_manager.h"
#include "transaction/transaction_manager.h"

const std::string TEST_DB_NAME = "LockManagerTestDB";

class LockManagerTest : public ::testing::Test {
   public:
    std::unique_ptr<LockManager> lock_manager_;
    std::unique_ptr<DiskManager> disk_manager_;
    std::unique_ptr<BufferPoolManager> buffer_pool_manager_;
    std::unique_ptr<RmManager> rm_manager_;
    std::unique_ptr<IxManager> ix_manager_;
    std::unique_ptr<SmManager> sm_manager_;
    std::unique_ptr<LogManager> log_manager_;
    std::unique_ptr<TransactionManager> txn_manager_;

   public:
    void SetUp() override {
        ::testing::Test::SetUp();
        lock_manager_ = std::make_unique<LockManager>();
        disk_manager_ = std::make_unique<DiskManager>();
        buffer_pool_manager_ = std::make_unique<BufferPoolManager>(
            BUFFER_POOL_SIZE, disk_manager_.get());
        rm_manager_ = std::make_unique<RmManager>(disk_manager_.get(),
                                                  buffer_pool_manager_.get());
        ix_manager_ = std::make_unique<IxManager>(disk_manager_.get(),
                                                  buffer_pool_manager_.get());
        sm_manager_ = std::make_unique<SmManager>(
            disk_manager_.get(), buffer_pool_manager_.get(), rm_manager_.get(),
            ix_manager_.get());
        log_manager_ = std::make_unique<LogManager>(disk_manager_.get());
        txn_manager_ = std::make_unique<TransactionManager>(lock_manager_.get(),
                                                            sm_manager_.get());
    }

    // 清理函数，每个测试用例运行结束后调用
    void TearDown() override {
        // 清理管理器实例
        lock_manager_.reset();
        disk_manager_.reset();
        buffer_pool_manager_.reset();
        rm_manager_.reset();
        ix_manager_.reset();
        sm_manager_.reset();
        log_manager_.reset();
        txn_manager_.reset();
        ::testing::Test::TearDown();
    }

    void init_txn_and_rows(std::vector<Transaction*>& txns,
                           std::vector<Rid>& rows, int num) {
        for (int i = 0; i < num; ++i) {
            rows.push_back({0, i});
            Transaction* t = new Transaction(i);
            txns.emplace_back(t);
            EXPECT_EQ(i, txns[i]->get_transaction_id());  // 验证事务ID
        }
    }

    void clear_txns(std::vector<Transaction*>& txns) {
        for (Transaction* txn : txns) {
            delete txn;
        }
        txns.clear();
    }

    void commit_txn(Transaction* txn) {
        txn_manager_->commit(txn, log_manager_.get());
        EXPECT_EQ(txn->get_state(), TransactionState::COMMITTED);
    }

    void abort_txn(Transaction* txn) {
        txn_manager_->abort(txn, log_manager_.get());
        EXPECT_EQ(txn->get_state(), TransactionState::ABORTED);
    }

    void s_lock_on_rows(Transaction* txn, const std::vector<Rid>& rows) {
        for (const Rid& row : rows) {
            bool res = lock_manager_->lock_shared_on_record(txn, row, 0);
            spdlog::info("txn_id : {}, Lock_Row S row_id: ({})",
                         txn->get_transaction_id(), row.slot_no);
            std::this_thread::sleep_for(
                std::chrono::milliseconds(rand() % 200));
            EXPECT_TRUE(res);
            EXPECT_EQ(txn->get_state(), TransactionState::GROWING);
        }
    }

    void s_unlock_on_rows(Transaction* txn, const std::vector<Rid>& rows) {
        for (const Rid& row : rows) {
            LockDataId lid(0, row, LockDataType::RECORD);
            bool res = lock_manager_->unlock(txn, lid);
            spdlog::info("txn_id : {}, unlock_row S row_id: ({})",
                         txn->get_transaction_id(), row.slot_no);
            EXPECT_TRUE(res);
            EXPECT_EQ(txn->get_state(), TransactionState::SHRINKING);
        }
    }

    void x_lock_on_rows(Transaction* txn, const std::vector<Rid>& rows) {
        for (const Rid& row : rows) {
            bool res = lock_manager_->lock_exclusive_on_record(txn, row, 0);
            spdlog::info("txn_id : {}, Lock_Row X row_id: ({})",
                         txn->get_transaction_id(), row.slot_no);
            std::this_thread::sleep_for(
                std::chrono::milliseconds(rand() % 200));
            EXPECT_TRUE(res);
            EXPECT_EQ(txn->get_state(), TransactionState::GROWING);
        }
    }

    void x_unlock_on_rows(Transaction* txn, const std::vector<Rid>& rows) {
        for (const Rid& row : rows) {
            LockDataId lid(0, row, LockDataType::RECORD);
            bool res = lock_manager_->unlock(txn, lid);
            spdlog::info("txn_id : {}, unlock_row X row_id: ({})",
                         txn->get_transaction_id(), row.slot_no);
            EXPECT_TRUE(res);
            EXPECT_EQ(txn->get_state(), TransactionState::SHRINKING);
        }
    }
};

TEST_F(LockManagerTest, BasicTest1_S_Lock) {
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 10;

    init_txn_and_rows(txns, rows_, num);

    auto task = [&](int txn_id) {
        std::this_thread::sleep_for(std::chrono::milliseconds(rand() % 500));

        bool res;

        res = lock_manager_->lock_IS_on_table(txns[txn_id], 0);
        EXPECT_TRUE(res);
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::GROWING);

        s_lock_on_rows(txns[txn_id], rows_);

        s_unlock_on_rows(txns[txn_id], rows_);

        LockDataId lid(0, LockDataType::TABLE);
        res = lock_manager_->unlock(txns[txn_id], lid);
        EXPECT_TRUE(res);
        std::cout << "txn_id : " << txn_id << "UnLockTable o_id: 0"
                  << std::endl;
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::SHRINKING);

        commit_txn(txns[txn_id]);
    };

    std::vector<std::thread> threads;
    threads.reserve(num);

    for (int i = 0; i < num; ++i) {
        threads.emplace_back(std::thread{task, i});
    }

    for (auto& thread : threads) {
        thread.join();
    }

    clear_txns(txns);
}

TEST_F(LockManagerTest, BasicTest1_X_Lock) {
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 10;

    for (int i = 0; i < num; ++i) {
        rows_.push_back(Rid{0, i});
        Transaction* t = new Transaction(i);
        txns.emplace_back(t);
        EXPECT_EQ(i, txns[i]->get_transaction_id());
    }

    auto task = [&](int txn_id) {
        std::this_thread::sleep_for(std::chrono::milliseconds(rand() % 500));

        bool res;

        res = lock_manager_->lock_IX_on_table(txns[txn_id], 0);
        EXPECT_TRUE(res);
        std::cout << "txn_id : " << txn_id << " LockTable IX o_id: 0";
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::GROWING);

        x_lock_on_rows(txns[txn_id], rows_);

        x_unlock_on_rows(txns[txn_id], rows_);
        LockDataId lid(0, LockDataType::TABLE);
        res = lock_manager_->unlock(txns[txn_id], lid);
        EXPECT_TRUE(res);
        std::cout << "txn_id : " << txn_id << "UnLockTable o_id: 0"
                  << std::endl;
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::SHRINKING);

        commit_txn(txns[txn_id]);
    };

    std::vector<std::thread> threads;
    threads.reserve(num);

    for (int i = 0; i < num; ++i) {
        threads.emplace_back(std::thread{task, i});
    }

    for (auto& thread : threads) {
        thread.join();
    }

    clear_txns(txns);
}

TEST_F(LockManagerTest, ConcurrentLocking) {
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 10;

    init_txn_and_rows(txns, rows_, num);

    auto task = [&](int txn_id) {
        bool res;
        res = lock_manager_->lock_IS_on_table(txns[txn_id], 0);
        EXPECT_TRUE(res);
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::GROWING);

        if (txn_id % 2 == 0) {
            s_lock_on_rows(txns[txn_id], {rows_[txn_id]});
        } else {
            x_lock_on_rows(txns[txn_id], {rows_[txn_id]});
        }

        std::this_thread::sleep_for(std::chrono::milliseconds(rand() % 100));

        if (txn_id % 2 == 0) {
            s_unlock_on_rows(txns[txn_id], {rows_[txn_id]});
        } else {
            x_unlock_on_rows(txns[txn_id], {rows_[txn_id]});
        }

        LockDataId lid(0, LockDataType::TABLE);
        res = lock_manager_->unlock(txns[txn_id], lid);
        EXPECT_TRUE(res);
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::SHRINKING);

        commit_txn(txns[txn_id]);
    };

    std::vector<std::thread> threads;
    threads.reserve(num);

    for (int i = 0; i < num; ++i) {
        threads.emplace_back(std::thread{task, i});
    }

    for (auto& thread : threads) {
        thread.join();
    }

    clear_txns(txns);
}

TEST_F(LockManagerTest, LockUpgrade) {

    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 5;

    init_txn_and_rows(txns, rows_, num);

    Transaction* txn = txns[0];

    bool res = lock_manager_->lock_IS_on_table(txn, 0);
    EXPECT_TRUE(res);
    EXPECT_EQ(txn->get_state(), TransactionState::GROWING);

    s_lock_on_rows(txn, rows_);

    // 尝试升级为排他锁
    res = lock_manager_->lock_IX_on_table(txn, 0);
    EXPECT_TRUE(res);
    EXPECT_EQ(txn->get_state(), TransactionState::GROWING);

    x_lock_on_rows(txn, rows_);

    x_unlock_on_rows(txn, rows_);

    LockDataId lid(0, LockDataType::TABLE);
    res = lock_manager_->unlock(txn, lid);
    EXPECT_TRUE(res);
    EXPECT_EQ(txn->get_state(), TransactionState::SHRINKING);

    commit_txn(txn);

    clear_txns(txns);
}

TEST_F(LockManagerTest, ComplexTest1) {
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 5;

    init_txn_and_rows(txns, rows_, num);

    auto task = [&](int txn_id) {
        bool res;

        if (txn_id % 2 == 0) {
            spdlog::info("txn_id: {} attempting IS lock on table", txn_id);
            res = lock_manager_->lock_IS_on_table(txns[txn_id], 0);
            EXPECT_TRUE(res);
        } else {
            spdlog::info("txn_id: {} attempting IX lock on table", txn_id);
            res = lock_manager_->lock_IX_on_table(txns[txn_id], 0);
            EXPECT_TRUE(res);
        }
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::GROWING);

        if (txn_id == 0) {
            spdlog::info("txn_id: {} attempting S lock on row 0", txn_id);
            s_lock_on_rows(txns[txn_id], {rows_[0]});
            std::this_thread::sleep_for(std::chrono::milliseconds(50));
            spdlog::info("txn_id: {} attempting X lock on row 1", txn_id);
            x_lock_on_rows(txns[txn_id], {rows_[1]});
        } else if (txn_id == 1) {
            spdlog::info("txn_id: {} attempting X lock on row 0", txn_id);
            x_lock_on_rows(txns[txn_id], {rows_[0]});
            std::this_thread::sleep_for(std::chrono::milliseconds(50));
            spdlog::info("txn_id: {} attempting S lock on row 1", txn_id);
            s_lock_on_rows(txns[txn_id], {rows_[1]});
        } else {
            spdlog::info("txn_id: {} attempting mixed locks", txn_id);
            s_lock_on_rows(txns[txn_id], {rows_[txn_id % 2]});
            x_lock_on_rows(txns[txn_id], {rows_[(txn_id + 1) % 2]});
        }

        spdlog::info("txn_id: {} unlocking all rows", txn_id);
        s_unlock_on_rows(txns[txn_id], rows_);
        x_unlock_on_rows(txns[txn_id], rows_);

        LockDataId lid(0, LockDataType::TABLE);
        res = lock_manager_->unlock(txns[txn_id], lid);
        EXPECT_TRUE(res);
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::SHRINKING);

        commit_txn(txns[txn_id]);
    };

    std::vector<std::thread> threads;
    threads.reserve(num);

    for (int i = 0; i < num; ++i) {
        threads.emplace_back(std::thread{task, i});
    }

    for (auto& thread : threads) {
        thread.join();
    }

    clear_txns(txns);
}

TEST_F(LockManagerTest, ComplexTest2_WaitDie) {
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 4;

    init_txn_and_rows(txns, rows_, num);

    auto task = [&](int txn_id) {
        spdlog::info("txn_id: {} starting", txn_id);

        bool res;
        if (txn_id < 2) {
            spdlog::info("txn_id: {} attempting IS lock on table", txn_id);
            res = lock_manager_->lock_IS_on_table(txns[txn_id], 0);
            EXPECT_TRUE(res);
        } else {
            spdlog::info("txn_id: {} attempting IX lock on table", txn_id);
            res = lock_manager_->lock_IX_on_table(txns[txn_id], 0);
            EXPECT_TRUE(res);
        }

        std::this_thread::sleep_for(std::chrono::milliseconds(rand() % 100));

        if (txn_id == 0) {
            s_lock_on_rows(txns[txn_id], {rows_[0]});
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
            x_lock_on_rows(txns[txn_id], {rows_[1]});
        } else if (txn_id == 1) {
            s_lock_on_rows(txns[txn_id], {rows_[1]});
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
            x_lock_on_rows(txns[txn_id], {rows_[0]});
        } else if (txn_id == 2) {
            x_lock_on_rows(txns[txn_id], {rows_[1]});
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
            s_lock_on_rows(txns[txn_id], {rows_[0]});
        } else {
            x_lock_on_rows(txns[txn_id], {rows_[0]});
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
            s_lock_on_rows(txns[txn_id], {rows_[1]});
        }

        s_unlock_on_rows(txns[txn_id], rows_);
        x_unlock_on_rows(txns[txn_id], rows_);

        LockDataId lid(0, LockDataType::TABLE);
        res = lock_manager_->unlock(txns[txn_id], lid);
        EXPECT_TRUE(res);
        EXPECT_EQ(txns[txn_id]->get_state(), TransactionState::SHRINKING);

        commit_txn(txns[txn_id]);
    };

    std::vector<std::thread> threads;
    threads.reserve(num);

    for (int i = 0; i < num; ++i) {
        threads.emplace_back(std::thread{task, i});
    }

    for (auto& thread : threads) {
        thread.join();
    }

    clear_txns(txns);
}

TEST_F(LockManagerTest, TransactionAbort) {
    std::vector<Rid> rows_;
    std::vector<Transaction*> txns;
    int num = 1;

    init_txn_and_rows(txns, rows_, num);

    Transaction* txn = txns[0];

    bool res = lock_manager_->lock_IS_on_table(txn, 0);
    EXPECT_TRUE(res);
    EXPECT_EQ(txn->get_state(), TransactionState::GROWING);

    s_lock_on_rows(txn, rows_);

    // 中止事务
    abort_txn(txn);

    // 检查所有锁是否已释放
    for (const Rid& row : rows_) {
        LockDataId lid(0, row, LockDataType::RECORD);
        EXPECT_FALSE(txn->is_row_s_locked(lid));
    }

    LockDataId lid(0, LockDataType::TABLE);
    EXPECT_FALSE(txn->is_tab_is_locked(lid));

    clear_txns(txns);
}
