/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"
#include "record/rm_file_handle.h"
#include "system/sm_manager.h"
#include "transaction/transaction.h"

std::unordered_map<txn_id_t, Transaction*> TransactionManager::txn_map = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction* TransactionManager::begin(Transaction* txn,
                                       LogManager* log_manager) {
    // 1. 判断传入事务参数是否为空指针
    // 2. 如果为空指针，创建新事务
    // 3. 把开始事务加入到全局事务表中
    // 4. 返回当前事务指针

    std::scoped_lock lock(latch_);

    if (txn == nullptr) {
        txn = new Transaction(next_txn_id_);
        txn->set_start_ts(next_timestamp_);
        next_txn_id_++, next_timestamp_++;
    }

    if (enable_logging) {
        BeginLogRecord log_record = BeginLogRecord(txn->get_transaction_id());
        log_record.prev_lsn_ = txn->get_prev_lsn();
        lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
        txn->set_prev_lsn(prev_lsn);
    }

    txn_map[txn->get_transaction_id()] = txn;
    return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
    // 1. 如果存在未提交的写操作，提交所有的写操作
    // 2. 释放所有锁
    // 3. 释放事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态

    std::scoped_lock lock(latch_);

    if (txn == nullptr) {
        return;
    }

    if (enable_logging) {
        CommitLogRecord log_record = CommitLogRecord(txn->get_transaction_id());
        log_record.prev_lsn_ = txn->get_prev_lsn();
        lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
        txn->set_prev_lsn(prev_lsn);
    }

    log_manager->flush_log_to_disk();

    auto write_set = txn->get_write_set();

    while (!write_set->empty()) {
        // TODO: may add logic delete here, a optimization method
        write_set->pop_back();
    }

    //  unlock all locks
    lock_manager_->unlock_all_locks(txn);

    txn->set_state(TransactionState::COMMITTED);
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction* txn, LogManager* log_manager) {
    // 1. 回滚所有写操作
    // 2. 释放所有锁
    // 3. 清空事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    std::scoped_lock lock{latch_};

    if (txn == nullptr) {
        return;
    }

    auto write_set = txn->get_write_set();

    // write_set 这个数据结构需要在 对数据有写操作时 进行维护
    while (!write_set->empty()) {
        auto write_record = write_set->back();
        write_record_rollback(write_record, log_manager, txn);
        write_set->pop_back();
    }

    //  unlock all locks
    lock_manager_->unlock_all_locks(txn);

    if (enable_logging) {
        AbortLogRecord log_record = AbortLogRecord(txn->get_transaction_id());
        log_record.prev_lsn_ = txn->get_prev_lsn();
        txn->set_prev_lsn(log_manager->add_log_to_buffer(&log_record));
    }

    log_manager->flush_log_to_disk();

    txn->set_state(TransactionState::ABORTED);
}

void TransactionManager::write_record_rollback(WriteRecord* write_record,
                                               LogManager* log_manager,
                                               Transaction* txn) {
    // 1. 写操作日志记录，写入日志文件，日志立即刷入磁盘
    // 2. 写操作提交，更新磁盘文件
    // 3. 更新事务状态

    auto cur_context =
        std::make_shared<Context>(lock_manager_, log_manager, txn);

    auto tab_name = write_record->GetTableName();
    auto& rid = write_record->GetRid();
    auto& record = write_record->GetRecord();

    switch (write_record->GetWriteType()) {
        case WType::INSERT_TUPLE: {
            rollback_insert(tab_name, rid, record, cur_context.get());
            break;
        }

        case WType::DELETE_TUPLE: {
            rollback_delete(tab_name, rid, record, cur_context.get());
            break;
        }

        case WType::UPDATE_TUPLE: {
            rollback_update(tab_name, rid, record, write_record->GetOldRecord(),
                            cur_context.get());
            break;
        }

        default:
            break;
    }
}

void TransactionManager::rollback_insert(const std::string& tab_name,
                                         const Rid& rid, const RmRecord& record,
                                         Context* context) {
    auto fh = sm_manager_->fhs_[tab_name].get();
    auto& tab_meta = sm_manager_->db_.get_table(tab_name);
    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(record.data);
        ih->delete_entry(key.get(), context->txn_);
    }
    fh->delete_record(rid, context);
}

void TransactionManager::rollback_delete(const std::string& tab_name,
                                         const Rid& rid, const RmRecord& record,
                                         Context* context) {
    auto fh = sm_manager_->fhs_[tab_name].get();
    auto& tab_meta = sm_manager_->db_.get_table(tab_name);
    // 先插入记录再插入索引
    fh->insert_record(rid, record.data, context);
    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(record.data);
        ih->insert_entry(key.get(), rid, context->txn_);
    }
}

void TransactionManager::rollback_update(const std::string& tab_name,
                                         const Rid& rid, const RmRecord& record,
                                         const RmRecord& old_record,
                                         Context* context) {
    auto fh = sm_manager_->fhs_[tab_name].get();
    auto& tab_meta = sm_manager_->db_.get_table(tab_name);
    // 删除新索引，用旧的记录反向更新，插入旧索引
    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(record.data);
        ih->delete_entry(key.get(), context->txn_);
    }

    fh->update_record(rid, old_record.data, context);

    for (auto& index : tab_meta.indexes) {
        auto ih = sm_manager_->ihs_
                      .at(sm_manager_->get_ix_manager()->get_index_name(
                          tab_name, index.cols))
                      .get();
        auto key = index.get_index_key(old_record.data);
        ih->insert_entry(key.get(), rid, context->txn_);
    }
}
